# Copyright (C) 2011 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from distutils.version import StrictVersion
import phonopy.structure.spglib as spg
from phonopy.structure.atoms import PhonopyAtoms


def get_supercell(unitcell, supercell_matrix, is_old_style=True, symprec=1e-5):
    return Supercell(unitcell,
                     supercell_matrix,
                     is_old_style=is_old_style,
                     symprec=symprec)


def get_primitive(supercell, primitive_frame, symprec=1e-5):
    return Primitive(supercell, primitive_frame, symprec=symprec)


def print_cell(cell, mapping=None, stars=None):
    symbols = cell.get_chemical_symbols()
    masses = cell.get_masses()
    magmoms = cell.get_magnetic_moments()
    lattice = cell.get_cell()
    print("Lattice vectors:")
    print("  a %20.15f %20.15f %20.15f" % tuple(lattice[0]))
    print("  b %20.15f %20.15f %20.15f" % tuple(lattice[1]))
    print("  c %20.15f %20.15f %20.15f" % tuple(lattice[2]))
    print("Atomic positions (fractional):")
    for i, v in enumerate(cell.get_scaled_positions()):
        num = " "
        if stars is not None:
            if i in stars:
                num = "*"
        num += "%d" % (i + 1)
        line = ("%5s %-2s%18.14f%18.14f%18.14f" %
                (num, symbols[i], v[0], v[1], v[2]))
        if masses is not None:
            line += " %7.3f" % masses[i]
        if magmoms is not None:
            line += "  %5.3f" % magmoms[i]
        if mapping is None:
            print(line)
        else:
            print(line + " > %d" % (mapping[i] + 1))


class Supercell(PhonopyAtoms):
    """Build supercell from supercell matrix and unit cell


    """

    def __init__(self,
                 unitcell,
                 supercell_matrix,
                 is_old_style=True,
                 symprec=1e-5):
        """

        Note
        ----

        ``is_old_style=True`` invokes the following algorithm.
        In this function, unit cell is considered
          [1,0,0]
          [0,1,0]
          [0,0,1].
        Supercell matrix is given by relative ratio, e.g,
          [-1, 1, 1]
          [ 1,-1, 1]  is for FCC from simple cubic.
          [ 1, 1,-1].
        In this case multiplicities of surrounding simple lattice are [2,2,2].
        First, create supercell with surrounding simple lattice.
        Second, trim the surrounding supercell with the target lattice.

        ``is_old_style=False`` calls the Smith normal form.

        These two algorithm may order atoms in different ways. So for the
        backward compatibitily, ``is_old_style=True`` is the default
        option. However the Smith normal form shows far better performance
        in case of very large supercell multiplicities.

        Parameters
        ----------
        unitcell: PhonopyAtoms
            Unit cell
        supercell_matrix: ndarray or list of list
            Transformation matrix from unit cell to supercell. The
            elements have to be integers.
            shape=(3,3)
        is_old_stype: bool
            This swithes the algorithms. See Note.
        symprec: float, optional
            Tolerance to find overlapping atoms in supercell cell. The default
            values is 1e-5.

        """

        self._is_old_style = is_old_style
        self._s2u_map = None
        self._u2s_map = None
        self._u2u_map = None
        self._supercell_matrix = np.array(supercell_matrix, dtype='intc')
        self._create_supercell(unitcell, symprec)

    def get_supercell_matrix(self):
        return self._supercell_matrix

    @property
    def s2u_map(self):
        return self._s2u_map

    def get_supercell_to_unitcell_map(self):
        return self.s2u_map

    @property
    def u2s_map(self):
        return self._u2s_map

    def get_unitcell_to_supercell_map(self):
        return self.u2s_map

    @property
    def u2u_map(self):
        return self._u2u_map

    def get_unitcell_to_unitcell_map(self):
        return self.u2u_map

    def _create_supercell(self, unitcell, symprec):
        mat = self._supercell_matrix

        if self._is_old_style:
            P = None
            multi = self._get_surrounding_frame(mat)
            # trim_fram is to trim overlapping atoms.
            trim_frame = np.array([mat[0] / float(multi[0]),
                                   mat[1] / float(multi[1]),
                                   mat[2] / float(multi[2])])
        else:
            # In the new style, it is unnecessary to trim atoms,
            if (np.diag(np.diagonal(mat)) != mat).any():
                snf = SNF3x3(mat)
                snf.run()
                P = snf.P
                multi = np.diagonal(snf.A)
            else:
                P = None
                multi = np.diagonal(mat)
            trim_frame = np.eye(3)

        sur_cell, u2sur_map = self._get_simple_supercell(unitcell, multi, P)
        trimmed_cell_ = _trim_cell(trim_frame, sur_cell, symprec)
        if trimmed_cell_:
            supercell, sur2s_map, mapping_table = trimmed_cell_
        else:
            return False

        num_satom = supercell.get_number_of_atoms()
        num_uatom = unitcell.get_number_of_atoms()
        N = num_satom // num_uatom

        if N != determinant(self._supercell_matrix):
            print("Supercell creation failed.")
            print("Probably some atoms are overwrapped. "
                  "The mapping table is give below.")
            print(mapping_table)
            PhonopyAtoms.__init__(self)
        else:
            PhonopyAtoms.__init__(
                self,
                numbers=supercell.get_atomic_numbers(),
                masses=supercell.get_masses(),
                magmoms=supercell.get_magnetic_moments(),
                scaled_positions=supercell.get_scaled_positions(),
                cell=supercell.get_cell(),
                pbc=True)
            self._u2s_map = np.arange(num_uatom) * N
            self._u2u_map = {j: i for i, j in enumerate(self._u2s_map)}
            self._s2u_map = np.array(u2sur_map)[sur2s_map] * N

    def _get_simple_supercell(self, unitcell, multi, P):
        if self._is_old_style:
            mat = np.diag(multi)
        else:
            mat = self._supercell_matrix

        # Scaled positions within the frame, i.e., create a supercell that
        # is made simply to multiply the input cell.
        positions = unitcell.get_scaled_positions()
        numbers = unitcell.get_atomic_numbers()
        masses = unitcell.get_masses()
        magmoms = unitcell.get_magnetic_moments()
        lattice = unitcell.get_cell()

        # Index of a axis runs fastest for creating lattice points.
        # See numpy.meshgrid document for the complicated index order for 3D
        b, c, a = np.meshgrid(range(multi[1]),
                              range(multi[2]),
                              range(multi[0]))
        lattice_points = np.c_[a.ravel(), b.ravel(), c.ravel()]

        if P is not None:
            # If supercell matrix is not a diagonal matrix,
            # Smith normal form is applied to find oblique basis vectors for
            # supercell and primitive cells, where their basis vectos are
            # parallel each other. By this reason, simple construction of
            # supercell becomes possible.
            P_inv = np.rint(np.linalg.inv(P)).astype(int)
            assert determinant(P_inv) == 1
            lattice_points = np.dot(lattice_points, P_inv.T)

        n = len(positions)
        n_l = len(lattice_points)
        # tile: repeat blocks
        # repeat: repeat each element
        positions_multi = np.dot(np.tile(lattice_points, (n, 1)) +
                                 np.repeat(positions, n_l, axis=0),
                                 np.linalg.inv(mat).T)
        numbers_multi = np.repeat(numbers, n_l)
        atom_map = np.repeat(np.arange(n), n_l)
        if masses is None:
            masses_multi = None
        else:
            masses_multi = np.repeat(masses, n_l)
        if magmoms is None:
            magmoms_multi = None
        else:
            magmoms_multi = np.repeat(magmoms, n_l)

        simple_supercell = PhonopyAtoms(numbers=numbers_multi,
                                        masses=masses_multi,
                                        magmoms=magmoms_multi,
                                        scaled_positions=positions_multi,
                                        cell=np.dot(mat, lattice),
                                        pbc=True)

        return simple_supercell, atom_map

    def _get_surrounding_frame(self, supercell_matrix):
        # Build a frame surrounding supercell lattice
        # For example,
        #  [2,0,0]
        #  [0,2,0] is the frame for FCC from simple cubic.
        #  [0,0,2]

        m = np.array(supercell_matrix)
        axes = np.array([[0, 0, 0],
                         m[:, 0],
                         m[:, 1],
                         m[:, 2],
                         m[:, 1] + m[:, 2],
                         m[:, 2] + m[:, 0],
                         m[:, 0] + m[:, 1],
                         m[:, 0] + m[:, 1] + m[:, 2]])
        frame = [max(axes[:, i]) - min(axes[:, i]) for i in (0, 1, 2)]
        return frame


class Primitive(PhonopyAtoms):
    """Primitive cell

    Attributes
    ----------
    primitive_matrix : ndarray
        Transformation matrix from supercell to primitive cell
        dtype='double'
        shape=(3,3)
    p2s_map : ndarray
        Mapping table from atoms in primitive cell to those in supercell.
        Supercell atomic indices are used.
        dtype='intc'
        shape=(num_atoms_in_primitive_cell,)
    s2p_map : ndarray
        Mapping table from atoms in supercell cell to those in primitive cell.
        Supercell atomic indices are used.
        dtype='intc'
        shape=(num_atoms_in_supercell,)
    p2p_map : dict
        Mapping of primitive cell atoms in supercell to those in primitive
        cell.
        ex. {0: 0, 4: 1}
    atomic_permutations : ndarray
        Atomic position permutation by pure translations is represented by
        changes of indices.
        dtype='intc'
        shape=(num_trans, num_atoms_in_supercell)
        ex.       supercell atomic indices
                 [[0, 1, 2, 3, 4, 5, 6, 7],
           trans  [1, 2, 3, 0, 5, 6, 7, 4],
          indices [2, 3, 0, 1, 6, 7, 4, 5],
                  [3, 0, 1, 2, 7, 4, 5, 6]]

    """

    def __init__(self, supercell, primitive_matrix, symprec=1e-5):
        """

        Parameters
        ----------
        supercell: PhonopyAtoms
            Supercell
        primitive_matrix: list of list or ndarray
            Transformation matrix to transform supercell to primitive cell
            such as:
               np.dot(primitive_matrix.T, supercell.get_cell())
            shape=(3,3)
        symprec: float, optional
            Tolerance to find overlapping atoms in primitive cell. The default
            values is 1e-5.

        """

        self._primitive_matrix = np.array(primitive_matrix,
                                          dtype='double', order='C')
        self._symprec = symprec
        self._p2s_map = None
        self._s2p_map = None
        self._p2p_map = None
        self._smallest_vectors = None
        self._multiplicity = None
        self._atomic_permutations = None
        self._primitive_cell(supercell)
        self._map_atomic_indices(supercell.get_scaled_positions())
        self._set_smallest_vectors(supercell)
        self._set_atomic_permutations(supercell)

    @property
    def primitive_matrix(self):
        return self._primitive_matrix

    def get_primitive_matrix(self):
        return self.primitive_matrix

    @property
    def p2s_map(self):
        return self._p2s_map

    def get_primitive_to_supercell_map(self):
        return self.p2s_map

    @property
    def s2p_map(self):
        return self._s2p_map

    def get_supercell_to_primitive_map(self):
        return self.s2p_map

    @property
    def p2p_map(self):
        return self._p2p_map

    def get_primitive_to_primitive_map(self):
        return self.p2p_map

    def get_smallest_vectors(self):
        return self._smallest_vectors, self._multiplicity

    @property
    def atomic_permutations(self):
        return self._atomic_permutations

    def get_atomic_permutations(self):
        return self.atomic_permutations

    def _primitive_cell(self, supercell):
        trimmed_cell_ = _trim_cell(self._primitive_matrix,
                                   supercell,
                                   self._symprec)
        if trimmed_cell_:
            trimmed_cell, p2s_map, mapping_table = trimmed_cell_
            PhonopyAtoms.__init__(
                self,
                numbers=trimmed_cell.get_atomic_numbers(),
                masses=trimmed_cell.get_masses(),
                magmoms=trimmed_cell.get_magnetic_moments(),
                scaled_positions=trimmed_cell.get_scaled_positions(),
                cell=trimmed_cell.get_cell(),
                pbc=True)
            self._p2s_map = np.array(p2s_map, dtype='intc')
        else:
            raise ValueError

    def _map_atomic_indices(self, s_pos_orig):
        frac_pos = np.dot(s_pos_orig, np.linalg.inv(self._primitive_matrix).T)

        p2s_positions = frac_pos[self._p2s_map]
        s2p_map = []
        for s_pos in frac_pos:

            # Compute distances from s_pos to all positions in _p2s_map.
            frac_diffs = p2s_positions - s_pos
            frac_diffs -= np.rint(frac_diffs)

            cart_diffs = np.dot(frac_diffs, self.cell)
            distances = np.sqrt((cart_diffs**2).sum(axis=1))

            # Find the one distance that's short enough.
            matches = (distances < self._symprec).tolist()
            assert matches.count(True) == 1
            index = matches.index(True)

            s2p_map.append(self._p2s_map[index])

        self._s2p_map = np.array(s2p_map, dtype='intc')
        self._p2p_map = dict([(j, i) for i, j in enumerate(self._p2s_map)])

    def _set_smallest_vectors(self, supercell):
        self._smallest_vectors, self._multiplicity = _get_smallest_vectors(
            supercell, self, symprec=self._symprec)

    def _set_atomic_permutations(self, supercell):
        positions = supercell.get_scaled_positions()
        diff = positions - positions[self._p2s_map[0]]
        trans = np.array(diff[np.where(self._s2p_map == self._p2s_map[0])[0]],
                         dtype='double', order='C')
        rotations = np.array([np.eye(3, dtype='intc')] * len(trans),
                             dtype='intc', order='C')
        self._atomic_permutations = compute_all_sg_permutations(
            positions,
            rotations,
            trans,
            np.array(supercell.get_cell().T, dtype='double', order='C'),
            self._symprec)


def _trim_cell(relative_axes, cell, symprec):
    """Trim overlapping atoms

    Parameters
    ----------
    relative_axes: ndarray
        Transformation matrix to transform supercell to a smaller cell such as:
            trimmed_lattice = np.dot(relative_axes.T, cell.get_cell())
        shape=(3,3)
    cell: PhonopyAtoms
        A supercell
    symprec: float
        Tolerance to find overlapping atoms in the trimmed cell

    Returns
    -------
    tuple
        trimmed_cell, extracted_atoms, mapping_table

    """
    positions = cell.get_scaled_positions()
    numbers = cell.get_atomic_numbers()
    masses = cell.get_masses()
    magmoms = cell.get_magnetic_moments()
    lattice = cell.get_cell()
    trimmed_lattice = np.dot(relative_axes.T, lattice)

    trimmed_positions = []
    trimmed_numbers = []
    if masses is None:
        trimmed_masses = None
    else:
        trimmed_masses = []
    if magmoms is None:
        trimmed_magmoms = None
    else:
        trimmed_magmoms = []
    extracted_atoms = []

    positions_in_new_lattice = np.dot(positions,
                                      np.linalg.inv(relative_axes).T)
    positions_in_new_lattice -= np.floor(positions_in_new_lattice)
    trimmed_positions = np.zeros_like(positions_in_new_lattice)
    num_atom = 0

    mapping_table = np.arange(len(positions), dtype='intc')
    for i, pos in enumerate(positions_in_new_lattice):
        is_overlap = False
        if num_atom > 0:
            diff = trimmed_positions[:num_atom] - pos
            diff -= np.rint(diff)
            # Older numpy doesn't support axis argument.
            # distances = np.linalg.norm(np.dot(diff, trimmed_lattice), axis=1)
            # overlap_indices = np.where(distances < symprec)[0]
            distances = np.sqrt(
                np.sum(np.dot(diff, trimmed_lattice) ** 2, axis=1))
            overlap_indices = np.where(distances < symprec)[0]
            if len(overlap_indices) > 0:
                assert len(overlap_indices) == 1
                is_overlap = True
                mapping_table[i] = extracted_atoms[overlap_indices[0]]

        if not is_overlap:
            trimmed_positions[num_atom] = pos
            num_atom += 1
            trimmed_numbers.append(numbers[i])
            if masses is not None:
                trimmed_masses.append(masses[i])
            if magmoms is not None:
                trimmed_magmoms.append(magmoms[i])
            extracted_atoms.append(i)

    # scale is not always to become integer.
    scale = 1.0 / np.linalg.det(relative_axes)
    if len(numbers) == np.rint(scale * len(trimmed_numbers)):
        trimmed_cell = PhonopyAtoms(
            numbers=trimmed_numbers,
            masses=trimmed_masses,
            magmoms=trimmed_magmoms,
            scaled_positions=trimmed_positions[:num_atom],
            cell=trimmed_lattice,
            pbc=True)
        return trimmed_cell, extracted_atoms, mapping_table
    else:
        return False


#
# Delaunay and Niggli reductions
#
def get_reduced_bases(lattice,
                      method='niggli',
                      tolerance=1e-5):
    """Search kinds of shortest basis vectors

    Parameters
    ----------
    lattice : ndarray or list of list
        Basis vectors by row vectors, [a, b, c]^T
        shape=(3, 3)
    method : str
        delaunay: Delaunay reduction
        niggli: Niggli reduction
    tolerance : float
        Tolerance to find shortest basis vecotrs

    Returns
    --------
    Reduced basis as row vectors, [a_red, b_red, c_red]^T
        dtype='double'
        shape=(3, 3)
        order='C'

    """

    if method == 'niggli':
        return spg.niggli_reduce(lattice, eps=tolerance)
    else:
        return spg.delaunay_reduce(lattice, eps=tolerance)


def _get_smallest_vectors(supercell, primitive, symprec=1e-5):
    p2s_map = primitive.p2s_map
    supercell_pos = supercell.scaled_positions
    primitive_pos = supercell_pos[p2s_map]
    supercell_bases = supercell.cell
    primitive_bases = primitive.cell
    svecs, multi = get_smallest_vectors(
        supercell_bases, supercell_pos, primitive_pos, symprec=symprec)
    trans_mat_float = np.dot(supercell_bases, np.linalg.inv(primitive_bases))
    trans_mat = np.rint(trans_mat_float).astype(int)
    assert (np.abs(trans_mat_float - trans_mat) < 1e-8).all()
    svecs = np.array(np.dot(svecs, trans_mat), dtype='double', order='C')
    return svecs, multi


def get_smallest_vectors(supercell_bases,
                         supercell_pos,
                         primitive_pos,
                         symprec=1e-5):
    """Find shortest atomic pair vectors

    Note
    ----
    Shortest vectors from an atom in primitive cell to an atom in
    supercell in the fractional coordinates of primitive cell. If an
    atom in supercell is on the border centered at an atom in
    primitive and there are multiple vectors that have the same
    distance (up to tolerance) and different directions, several
    shortest vectors are stored.
    In fact, this method is not limited to search shortest vectors between
    sueprcell atoms and primitive cell atoms, but can be used to measure
    shortest vectors between atoms in periodic supercell lattice frame.

    Parameters
    ----------
    supercell_bases : ndarray
        Supercell basis vectors as row vectors, (a, b, c)^T. Must be order='C'.
        dtype='double'
        shape=(3, 3)
    supercell_pos : array_like
        Atomic positions in fractional coordinates of supercell.
        dtype='double'
        shape=(size_super, 3)
    primitive_pos : array_like
        Atomic positions in fractional coordinates of supercell. Note that not
        in fractional coodinates of primitive cell.
        dtype='double'
        shape=(size_prim, 3)
    symprec : float, optional, default=1e-5
        Tolerance to find equal distances of vectors

    Returns
    -------
    shortest_vectors : ndarray
        Shortest vectors in supercell coordinates. The 27 in shape is the
        possible maximum number of elements.
        dtype='double'
        shape=(size_super, size_prim, 27, 3)
    multiplicities : ndarray
        Number of equidistance shortest vectors
        dtype='intc'
        shape=(size_super, size_prim)

    """

    reduced_cell_method = 'niggli'
    reduced_bases = get_reduced_bases(supercell_bases,
                                      method=reduced_cell_method,
                                      tolerance=symprec)
    trans_mat_float = np.dot(supercell_bases, np.linalg.inv(reduced_bases))
    trans_mat = np.rint(trans_mat_float).astype(int)
    assert (np.abs(trans_mat_float - trans_mat) < 1e-8).all()
    trans_mat_inv_float = np.linalg.inv(trans_mat)
    trans_mat_inv = np.rint(trans_mat_inv_float).astype(int)
    assert (np.abs(trans_mat_inv_float - trans_mat_inv) < 1e-8).all()

    # Reduce all positions into the cell formed by the reduced bases.
    supercell_fracs = np.dot(supercell_pos, trans_mat)
    supercell_fracs -= np.rint(supercell_fracs)
    supercell_fracs = np.array(supercell_fracs, dtype='double', order='C')
    primitive_fracs = np.dot(primitive_pos, trans_mat)
    primitive_fracs -= np.rint(primitive_fracs)
    primitive_fracs = np.array(primitive_fracs, dtype='double', order='C')

    # For each vector, we will need to consider all nearby images in the
    # reduced bases. The lattice points at which supercell images are searched
    # are composed by linear combinations of three vectors in
    # (0, a, b, c, -a-b-c, -a, -b, -c, a+b+c). There are finally 65 lattice
    # points. There is no proof that this is enough.
    lattice_1D = (-1, 0, 1)
    lattice_4D = np.array([[i, j, k, l]
                           for i in lattice_1D
                           for j in lattice_1D
                           for k in lattice_1D
                           for l in lattice_1D],
                          dtype='intc', order='C')
    bases = [[1, 0, 0], [0, 1, 0], [0, 0, 1], [-1, -1, -1]]
    lattice_points = np.dot(lattice_4D, bases)
    if StrictVersion(np.__version__) >= StrictVersion('1.13.0'):
        lattice_points = np.array(np.unique(lattice_points, axis=0),
                                  dtype='intc', order='C')
    else:
        unique_indices = np.unique([
            np.nonzero(np.abs(lattice_points - point).sum(axis=1) == 0)[0][0]
            for point in lattice_points])
        lattice_points = np.array(lattice_points[unique_indices],
                                  dtype='intc', order='C')

    # This shortest_vectors is already used at many locations.
    # Therefore the constant number 27 = 3*3*3 can not be easily changed.
    shortest_vectors = np.zeros(
        (len(supercell_fracs), len(primitive_fracs), 27, 3),
        dtype='double', order='C')
    multiplicity = np.zeros((len(supercell_fracs), len(primitive_fracs)),
                            dtype='intc', order='C')
    import phonopy._phonopy as phonoc
    phonoc.gsv_set_smallest_vectors(
        shortest_vectors,
        multiplicity,
        supercell_fracs,
        primitive_fracs,
        lattice_points,
        np.array(reduced_bases.T, dtype='double', order='C'),
        np.array(trans_mat_inv.T, dtype='intc', order='C'),
        symprec)

    # Here's where things get interesting.
    # We want to avoid manually iterating over all possible pairings of
    # supercell atoms and primitive atoms, because doing so creates a
    # tight loop in larger structures that is difficult to optimize.
    #
    # Furthermore, it seems wise to call numpy.dot on as large of an array
    # as possible, since numpy can shell out to BLAS to handle the
    # real heavy lifting.

    # lattice_1D = (-1, 0, 1)
    # lattice_points = np.array([[i, j, k]
    #                            for i in lattice_1D
    #                            for j in lattice_1D
    #                            for k in lattice_1D],
    #                           dtype='intc', order='C')

    # # For every atom in the supercell and every atom in the primitive cell,
    # # we want 27 images of the vector between them.
    # #
    # # 'None' is used to insert trivial axes to make these arrays broadcast.
    # #
    # # shape: (size_super, size_prim, 27, 3)
    # candidate_fracs = (
    #     supercell_fracs[:, None, None, :]    # shape: (size_super, 1, 1, 3)
    #     - primitive_fracs[None, :, None, :]  # shape: (1, size_prim, 1, 3)
    #     + lattice_points[None, None, :, :]   # shape: (1, 1, 27, 3)
    # )

    # # To compute the lengths, we want cartesian positions.
    # #
    # # Conveniently, calling 'numpy.dot' between a 4D array and a 2D array
    # # does vector-matrix multiplication on each row vector in the last axis
    # # of the 4D array.
    # #
    # # shape: (size_super, size_prim, 27)
    # lengths = np.array(np.sqrt(
    #     np.sum(np.dot(candidate_fracs, reduced_bases)**2, axis=-1)),
    #                    dtype='double', order='C')

    # # Create the output, initially consisting of all candidate vectors scaled
    # # by the primitive cell.
    # #
    # # shape: (size_super, size_prim, 27, 3)
    # candidate_vectors = np.array(np.dot(candidate_fracs, trans_mat_inv),
    #                              dtype='double', order='C')

    # # The last final bits are done in C.
    # #
    # # We will gather the shortest ones from each list of 27 vectors.
    # shortest_vectors = np.zeros_like(candidate_vectors,
    #                                  dtype='double', order='C')
    # multiplicity = np.zeros(shortest_vectors.shape[:2], dtype='intc',
    #                         order='C')

    # import phonopy._phonopy as phonoc
    # phonoc.gsv_copy_smallest_vectors(shortest_vectors,
    #                                  multiplicity,
    #                                  candidate_vectors,
    #                                  lengths,
    #                                  symprec)

    return shortest_vectors, multiplicity


def compute_all_sg_permutations(positions,  # scaled positions
                                rotations,  # scaled
                                translations,  # scaled
                                lattice,  # column vectors
                                symprec):
    """Compute a permutation for every space group operation.

    See 'compute_permutation_for_rotation' for more info.

    Output has shape (num_rot, num_pos)

    """

    out = []  # Finally the shape is fixed as (num_sym, num_pos_of_supercell).
    for (sym, t) in zip(rotations, translations):
        rotated_positions = np.dot(positions, sym.T) + t
        out.append(compute_permutation_for_rotation(positions,
                                                    rotated_positions,
                                                    lattice,
                                                    symprec))
    return np.array(out, dtype='intc', order='C')


def compute_permutation_for_rotation(positions_a,  # scaled positions
                                     positions_b,
                                     lattice,  # column vectors
                                     symprec):
    """Get the overall permutation such that

        positions_a[perm[i]] == positions_b[i]   (modulo the lattice)

    or in numpy speak,

        positions_a[perm] == positions_b   (modulo the lattice)

    This version is optimized for the case where positions_a and positions_b
    are related by a rotation.

    """

    # Sort both sides by some measure which is likely to produce a small
    # maximum value of (sorted_rotated_index - sorted_original_index).
    # The C code is optimized for this case, reducing an O(n^2)
    # search down to ~O(n). (for O(n log n) work overall, including the sort)
    #
    # We choose distance from the nearest bravais lattice point as our measure.
    def sort_by_lattice_distance(fracs):
        carts = np.dot(fracs - np.rint(fracs), lattice.T)
        perm = np.argsort(np.sum(carts**2, axis=1))
        sorted_fracs = np.array(fracs[perm], dtype='double', order='C')
        return perm, sorted_fracs

    (perm_a, sorted_a) = sort_by_lattice_distance(positions_a)
    (perm_b, sorted_b) = sort_by_lattice_distance(positions_b)

    # Call the C code on our conditioned inputs.
    perm_between = _compute_permutation_c(sorted_a,
                                          sorted_b,
                                          lattice,
                                          symprec)

    # Compose all of the permutations for the full permutation.
    #
    # Note the following properties of permutation arrays:
    #
    # 1. Inverse:         if  x[perm] == y  then  x == y[argsort(perm)]
    # 2. Associativity:   x[p][q] == x[p[q]]
    return perm_a[perm_between][np.argsort(perm_b)]


def _compute_permutation_c(positions_a,  # scaled positions
                           positions_b,
                           lattice,  # column vectors
                           symprec):
    """Version of '_compute_permutation_for_rotation' which just directly
    calls the C function, without any conditioning of the data.
    Skipping the conditioning step makes this EXTREMELY slow on large
    structures.

    """

    permutation = np.zeros(shape=(len(positions_a),), dtype='intc')

    def permutation_error():
        raise ValueError("Input forces are not enough to calculate force constants, "
                         "or something wrong (e.g. crystal structure does not match).")

    try:
        import phonopy._phonopy as phonoc
        tolerance = symprec
        for _ in range(20):
            is_found = phonoc.compute_permutation(permutation,
                                                  lattice,
                                                  positions_a,
                                                  positions_b,
                                                  tolerance)
            if is_found:
                break
            else:
                tolerance *= 1.05

        if tolerance / symprec > 1.5:
            import warnings
            msg = ("Crystal structure is distorted in a tricky way so that "
                   "phonopy could not handle the crystal symmetry properly. "
                   "It is recommended to symmetrize crystal structure well "
                   "and then re-start phonon calculation from scratch.")
            warnings.warn(msg)

        if not is_found:
            permutation_error()

    except ImportError:
        for i, pos_b in enumerate(positions_b):
            diffs = positions_a - pos_b
            diffs -= np.rint(diffs)
            diffs = np.dot(diffs, lattice.T)

            possible_j = np.nonzero(
                np.sqrt(np.sum(diffs**2, axis=1)) < symprec)[0]
            if len(possible_j) != 1:
                permutation_error()

            permutation[i] = possible_j[0]

        if -1 in permutation:
            permutation_error()

    return permutation


#
# Other tiny tools
#
def get_angles(lattice):
    a, b, c = get_cell_parameters(lattice)
    alpha = np.arccos(np.vdot(lattice[1], lattice[2]) / b / c) / np.pi * 180
    beta = np.arccos(np.vdot(lattice[2], lattice[0]) / c / a) / np.pi * 180
    gamma = np.arccos(np.vdot(lattice[0], lattice[1]) / a / b) / np.pi * 180
    return alpha, beta, gamma


def get_cell_parameters(lattice):
    return np.sqrt(np.dot(lattice, lattice.transpose()).diagonal())


def get_cell_matrix(a, b, c, alpha, beta, gamma):
    # These follow 'matrix_lattice_init' in matrix.c of GDIS
    alpha *= np.pi / 180
    beta *= np.pi / 180
    gamma *= np.pi / 180
    b1 = np.cos(gamma)
    b2 = np.sin(gamma)
    b3 = 0.0
    c1 = np.cos(beta)
    c2 = (2 * np.cos(alpha) + b1**2 + b2**2 - 2 * b1 * c1 - 1) / (2 * b2)
    c3 = np.sqrt(1 - c1**2 - c2**2)
    lattice = np.zeros((3, 3), dtype='double')
    lattice[0, 0] = a
    lattice[1] = np.array([b1, b2, b3]) * b
    lattice[2] = np.array([c1, c2, c3]) * c
    return lattice


def determinant(m):
    return (m[0][0] * m[1][1] * m[2][2] -
            m[0][0] * m[1][2] * m[2][1] +
            m[0][1] * m[1][2] * m[2][0] -
            m[0][1] * m[1][0] * m[2][2] +
            m[0][2] * m[1][0] * m[2][1] -
            m[0][2] * m[1][1] * m[2][0])


#
# Smith normal form for 3x3 integer matrix
# This code is maintained at https://github.com/atztogo/snf3x3.
#
class SNF3x3(object):
    def __init__(self, A):
        self._A_orig = np.array(A, dtype='intc')
        self._A = np.array(A, dtype='intc')
        self._Ps = []
        self._Qs = []
        self._L = []
        self._P = None
        self._Q = None
        self._attempt = 0

    def run(self):
        for i in self:
            pass

    def __iter__(self):
        return self

    def __next__(self):
        self._attempt += 1
        if self._first():
            if self._second():
                self._set_PQ()
                raise StopIteration
        return self._attempt

    def next(self):
        self.__next__()

    @property
    def A(self):
        return self._A.copy()

    @property
    def P(self):
        return self._P.copy()

    @property
    def Q(self):
        return self._Q.copy()

    def _set_PQ(self):
        if np.linalg.det(self._A) < 0:
            for i in range(3):
                if self._A[i, i] < 0:
                    self._flip_sign_row(i)
            self._Ps += self._L
            self._L = []

        P = np.eye(3, dtype='intc')
        for _P in self._Ps:
            P = np.dot(_P, P)
        Q = np.eye(3, dtype='intc')
        for _Q in self._Qs:
            Q = np.dot(Q, _Q.T)

        if np.linalg.det(P) < 0:
            P = -P
            Q = -Q

        self._P = P
        self._Q = Q

    def _first(self):
        self._first_one_loop()
        A = self._A
        if A[1, 0] == 0 and A[2, 0] == 0:
            return True
        elif A[1, 0] % A[0, 0] == 0 and A[2, 0] % A[0, 0] == 0:
            self._first_finalize()
            self._Ps += self._L
            self._L = []
            return True
        else:
            return False

    def _first_one_loop(self):
        self._first_column()
        self._Ps += self._L
        self._L = []
        self._A = self._A.T
        self._first_column()
        self._Qs += self._L
        self._L = []
        self._A = self._A.T

    def _first_column(self):
        i = self._search_first_pivot()
        if i > 0:
            self._swap_rows(0, i)

        if self._A[1, 0] != 0:
            self._zero_first_column(1)
        if self._A[2, 0] != 0:
            self._zero_first_column(2)

    def _zero_first_column(self, j):
        if self._A[j, 0] < 0:
            self._flip_sign_row(j)
        A = self._A
        r, s, t = xgcd([A[0, 0], A[j, 0]])
        self._set_zero(0, j, A[0, 0], A[j, 0], r, s, t)

    def _search_first_pivot(self):
        A = self._A
        for i in range(3):  # column index
            if A[i, 0] != 0:
                return i

    def _first_finalize(self):
        """Set zeros along the first colomn except for A[0, 0]

        This is possible only when A[1,0] and A[2,0] are dividable by A[0,0].

        """

        A = self._A
        L = np.eye(3, dtype='intc')
        L[1, 0] = -A[1, 0] // A[0, 0]
        L[2, 0] = -A[2, 0] // A[0, 0]
        self._L.append(L.copy())
        self._A = np.dot(L, self._A)

    def _second(self):
        """Find Smith normal form for Right-low 2x2 matrix"""

        self._second_one_loop()
        A = self._A
        if A[2, 1] == 0:
            return True
        elif A[2, 1] % A[1, 1] == 0:
            self._second_finalize()
            self._Ps += self._L
            self._L = []
            return True
        else:
            return False

    def _second_one_loop(self):
        self._second_column()
        self._Ps += self._L
        self._L = []
        self._A = self._A.T
        self._second_column()
        self._Qs += self._L
        self._L = []
        self._A = self._A.T

    def _second_column(self):
        """Right-low 2x2 matrix

        Assume elements in first row and column are all zero except for A[0,0].

        """

        if self._A[1, 1] == 0 and self._A[2, 1] != 0:
            self._swap_rows(1, 2)

        if self._A[2, 1] != 0:
            self._zero_second_column()

    def _zero_second_column(self):
        if self._A[2, 1] < 0:
            self._flip_sign_row(2)
        A = self._A
        r, s, t = xgcd([A[1, 1], A[2, 1]])
        self._set_zero(1, 2, A[1, 1], A[2, 1], r, s, t)

    def _second_finalize(self):
        """Set zero at A[2, 1]

        This is possible only when A[2,1] is dividable by A[1,1].

        """

        A = self._A
        L = np.eye(3, dtype='intc')
        L[2, 1] = -A[2, 1] // A[1, 1]
        self._L.append(L.copy())
        self._A = np.dot(L, self._A)

    def _swap_rows(self, i, j):
        """Swap i and j rows

        As the side effect, determinant flips.

        """

        L = np.eye(3, dtype='intc')
        L[i, i] = 0
        L[j, j] = 0
        L[i, j] = 1
        L[j, i] = 1
        self._L.append(L.copy())
        self._A = np.dot(L, self._A)

    def _flip_sign_row(self, i):
        """Multiply -1 for all elements in row"""

        L = np.eye(3, dtype='intc')
        L[i, i] = -1
        self._L.append(L.copy())
        self._A = np.dot(L, self._A)

    def _set_zero(self, i, j, a, b, r, s, t):
        """Let A[i, j] be zero based on Bezout's identity

           [ii ij]
           [ji jj] is a (k,k) minor of original 3x3 matrix.

        """

        L = np.eye(3, dtype='intc')
        L[i, i] = s
        L[i, j] = t
        L[j, i] = -b // r
        L[j, j] = a // r
        self._L.append(L.copy())
        self._A = np.dot(L, self._A)


def xgcd(vals):
    _xgcd = Xgcd(vals)
    return _xgcd.run()


class Xgcd(object):
    def __init__(self, vals):
        self._vals = np.array(vals, dtype='intc')

    def run(self):
        r0, r1 = self._vals
        s0 = 1
        s1 = 0
        t0 = 0
        t1 = 1
        for i in range(1000):
            r0, r1, s0, s1, t0, t1 = self._step(r0, r1, s0, s1, t0, t1)
            if r1 == 0:
                break
        self._rst = np.array([r0, s0, t0], dtype='intc')
        return self._rst

    def _step(self, r0, r1, s0, s1, t0, t1):
        q, m = divmod(r0, r1)
        r2 = m
        s2 = s0 - q * s1
        t2 = t0 - q * t1
        return r1, r2, s1, s2, t1, t2

    def __str__(self):
        v = self._vals
        r, s, t = self._rst
        return "%d = %d * (%d) + %d * (%d)" % (r, v[0], s, v[1], t)


def get_primitive_matrix_by_centring(centring):
    if centring == 'P':
        return [[1, 0, 0],
                [0, 1, 0],
                [0, 0, 1]]
    elif centring == 'F':
        return [[0, 1./2, 1./2],
                [1./2, 0, 1./2],
                [1./2, 1./2, 0]]
    elif centring == 'I':
        return [[-1./2, 1./2, 1./2],
                [1./2, -1./2, 1./2],
                [1./2, 1./2, -1./2]]
    elif centring == 'A':
        return [[1, 0, 0],
                [0, 1./2, -1./2],
                [0, 1./2, 1./2]]
    elif centring == 'C':
        return [[1./2, 1./2, 0],
                [-1./2, 1./2, 0],
                [0, 0, 1]]
    elif centring == 'R':
        return [[2./3, -1./3, -1./3],
                [1./3, 1./3, -2./3],
                [1./3, 1./3, 1./3]]
    else:
        return None


def guess_primitive_matrix(unitcell, symprec=1e-5):
    if unitcell.get_magnetic_moments() is not None:
        msg = "Can not be used with the unit cell having magnetic moments."
        raise RuntimeError(msg)

    lattice = unitcell.get_cell()
    cell = (lattice,
            unitcell.get_scaled_positions(),
            unitcell.get_atomic_numbers())
    dataset = spg.get_symmetry_dataset(cell, symprec=1e-5)
    tmat = dataset['transformation_matrix']
    centring = dataset['international'][0]
    pmat = get_primitive_matrix_by_centring(centring)
    return np.array(np.dot(np.linalg.inv(tmat), pmat),
                    dtype='double', order='C')


def estimate_supercell_matrix(spglib_dataset,
                              max_num_atoms=120):
    """Estimate supercell matrix from conventional cell

    Diagonal supercell matrix is estimated from basis vector lengths
    and maximum number of atoms to be accepted. Supercell is assumed
    to be made from the standardized cell and to be closest to sphere
    under keeping lattice symmetry. For triclinic, monoclinic, and
    orthorhombic cells, multiplicities for a, b, c are not constrained
    by symmetry. For tetragonal and hexagonal cells, multiplicities
    for a and b are chosen to be the same, and for cubic cell, those
    of a, b, c are the same.

    Parameters
    ----------
    spglib_dataset : tuple
        Spglib symmetry dataset
    max_num_atoms : int, optional
        Maximum number of atoms in created supercell to be tolerated.

    Returns
    -------
    list of three integer numbers
        Multiplicities for a, b, c basis vectors, respectively.

    """

    spg_num = spglib_dataset['number']
    num_atoms = len(spglib_dataset['std_types'])
    lengths = _get_lattice_parameters(spglib_dataset['std_lattice'])

    if spg_num <= 74:  # Triclinic, monoclinic, and orthorhombic
        multi = _get_multiplicity_abc(num_atoms, lengths, max_num_atoms)
    elif spg_num <= 194:  # Tetragonal and hexagonal
        multi = _get_multiplicity_ac(num_atoms, lengths, max_num_atoms)
    else:  # Cubic
        multi = _get_multiplicity_a(num_atoms, lengths, max_num_atoms)

    return multi


def _get_lattice_parameters(lattice):
    """Return basis vector lengths

    Parameters
    ----------
    lattice : array_like
        Basis vectors given as column vectors
        shape=(3, 3), dtype='double'

    Returns
    -------
    ndarray, shape=(3,), dtype='double'

    """

    return np.array(np.sqrt(np.dot(lattice.T, lattice).diagonal()),
                    dtype='double')


def _get_multiplicity_abc(num_atoms, lengths, max_num_atoms, max_iter=20):
    multi = [1, 1, 1]

    for i in range(max_iter):
        l_super = np.multiply(multi, lengths)
        min_index = np.argmin(l_super)
        multi[min_index] += 1
        if num_atoms * np.prod(multi) > max_num_atoms:
            multi[min_index] -= 1

    return multi


def _get_multiplicity_ac(num_atoms, lengths, max_num_atoms, max_iter=20):
    multi = [1, 1]
    a = lengths[0]
    c = lengths[2]

    for i in range(max_iter):
        l_super = np.multiply(multi, [a, c])
        min_index = np.argmin(l_super)
        multi[min_index] += 1
        if num_atoms * multi[0] ** 2 * multi[1] > max_num_atoms:
            multi[min_index] -= 1

    return [multi[0], multi[0], multi[1]]


def _get_multiplicity_a(num_atoms, lengths, max_num_atoms, max_iter=20):
    multi = 1
    for i in range(max_iter):
        multi += 1
        if num_atoms * multi ** 3 > max_num_atoms:
            multi -= 1

    return [multi, multi, multi]



# Copyright (C) 2011 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys
import numpy as np


def fracval(frac):
    if frac.find('/') == -1:
        return float(frac)
    else:
        x = frac.split('/')
        return float(x[0]) / float(x[1])


class Settings(object):
    def __init__(self):
        self._band_indices = None
        self._band_paths = None
        self._band_points = None
        self._cell_filename = None
        self._chemical_symbols = None
        self._cutoff_frequency = None
        self._displacement_distance = None
        self._dm_decimals = None
        self._fc_calculator = None
        self._fc_calculator_options = None
        self._fc_decimals = None
        self._fc_symmetry = False
        self._fpitch = None
        self._frequency_conversion_factor = None
        self._frequency_scale_factor = None
        self._gv_delta_q = None
        self._is_band_const_interval = False
        self._is_diagonal_displacement = True
        self._is_eigenvectors = False
        self._is_mesh_symmetry = True
        self._is_nac = False
        self._is_rotational_invariance = False
        self._is_plusminus_displacement = 'auto'
        self._is_symmetry = True
        self._is_tetrahedron_method = True
        self._is_time_reversal_symmetry = True
        self._is_trigonal_displacement = False
        self._magmoms = None
        self._masses = None
        self._mesh = None
        self._nac_method = None
        self._nac_q_direction = None
        self._num_frequency_points = None
        self._primitive_matrix = None
        self._qpoints = None
        self._read_qpoints = False
        self._sigma = None
        self._supercell_matrix = None
        self._symmetry_tolerance = None
        self._temperatures = None
        self._tmax = 1000
        self._tmin = 0
        self._tstep = 10

    def set_band_paths(self, band_paths):
        self._band_paths = band_paths

    def get_band_paths(self):
        return self._band_paths

    def set_band_points(self, band_points):
        self._band_points = band_points

    def get_band_points(self):
        return self._band_points

    def set_band_indices(self, band_indices):
        self._band_indices = band_indices

    def get_band_indices(self):
        return self._band_indices

    def set_cell_filename(self, cell_filename):
        self._cell_filename = cell_filename

    def get_cell_filename(self):
        return self._cell_filename

    def set_chemical_symbols(self, symbols):
        self._chemical_symbols = symbols

    def get_chemical_symbols(self):
        return self._chemical_symbols

    def set_cutoff_frequency(self, cutoff_frequency):
        self._cutoff_frequency = cutoff_frequency

    def get_cutoff_frequency(self):
        return self._cutoff_frequency

    def set_dm_decimals(self, decimals):
        self._dm_decimals = decimals

    def get_dm_decimals(self):
        return self._dm_decimals

    def set_displacement_distance(self, distance):
        self._displacement_distance = distance

    def get_displacement_distance(self):
        return self._displacement_distance

    def set_fc_calculator(self, fc_calculator):
        self._fc_calculator = fc_calculator

    def get_fc_calculator(self):
        return self._fc_calculator

    def set_fc_calculator_options(self, fc_calculator_options):
        self._fc_calculator_options = fc_calculator_options

    def get_fc_calculator_options(self):
        return self._fc_calculator_options

    def set_fc_symmetry(self, fc_symmetry):
        self._fc_symmetry = fc_symmetry

    def get_fc_symmetry(self):
        return self._fc_symmetry

    def set_fc_decimals(self, decimals):
        self._fc_decimals = decimals

    def get_fc_decimals(self):
        return self._fc_decimals

    def set_frequency_conversion_factor(self, frequency_conversion_factor):
        self._frequency_conversion_factor = frequency_conversion_factor

    def get_frequency_conversion_factor(self):
        return self._frequency_conversion_factor

    def set_frequency_pitch(self, fpitch):
        self._fpitch = fpitch

    def get_frequency_pitch(self):
        return self._fpitch

    def set_frequency_scale_factor(self, frequency_scale_factor):
        self._frequency_scale_factor = frequency_scale_factor

    def get_frequency_scale_factor(self):
        return self._frequency_scale_factor

    def set_num_frequency_points(self, num_frequency_points):
        self._num_frequency_points = num_frequency_points

    def get_num_frequency_points(self):
        return self._num_frequency_points

    def set_group_velocity_delta_q(self, gv_delta_q):
        self._gv_delta_q = gv_delta_q

    def get_group_velocity_delta_q(self):
        return self._gv_delta_q

    def set_is_band_const_interval(self, is_band_const_interval):
        self._is_band_const_interval = is_band_const_interval

    def get_is_band_const_interval(self):
        return self._is_band_const_interval

    def set_is_diagonal_displacement(self, is_diag):
        self._is_diagonal_displacement = is_diag

    def get_is_diagonal_displacement(self):
        return self._is_diagonal_displacement

    def set_is_eigenvectors(self, is_eigenvectors):
        self._is_eigenvectors = is_eigenvectors

    def get_is_eigenvectors(self):
        return self._is_eigenvectors

    def set_is_mesh_symmetry(self, is_mesh_symmetry):
        self._is_mesh_symmetry = is_mesh_symmetry

    def get_is_mesh_symmetry(self):
        return self._is_mesh_symmetry

    def set_is_nac(self, is_nac):
        self._is_nac = is_nac

    def get_is_nac(self):
        return self._is_nac

    def set_is_plusminus_displacement(self, is_pm):
        self._is_plusminus_displacement = is_pm

    def get_is_plusminus_displacement(self):
        return self._is_plusminus_displacement

    def set_is_rotational_invariance(self, is_rotational_invariance):
        self._is_rotational_invariance = is_rotational_invariance

    def get_is_rotational_invariance(self):
        return self._is_rotational_invariance

    def set_is_tetrahedron_method(self, is_thm):
        self._is_tetrahedron_method = is_thm

    def get_is_tetrahedron_method(self):
        return self._is_tetrahedron_method

    def set_is_trigonal_displacement(self, is_trigonal):
        self._is_trigonal_displacement = is_trigonal

    def get_is_trigonal_displacement(self):
        return self._is_trigonal_displacement

    def set_is_symmetry(self, is_symmetry):
        self._is_symmetry = is_symmetry

    def get_is_symmetry(self):
        return self._is_symmetry

    def set_magnetic_moments(self, magmoms):
        self._magmoms = magmoms

    def get_magnetic_moments(self):
        return self._magmoms

    def set_masses(self, masses):
        self._masses = masses

    def get_masses(self):
        return self._masses

    def set_max_temperature(self, tmax):
        self._tmax = tmax

    def get_max_temperature(self):
        return self._tmax

    def set_mesh_numbers(self, mesh):
        self._mesh = mesh

    def get_mesh_numbers(self):
        return self._mesh

    def set_min_temperature(self, tmin):
        self._tmin = tmin

    def get_min_temperature(self):
        return self._tmin

    def set_nac_method(self, nac_method):
        self._nac_method = nac_method

    def get_nac_method(self):
        return self._nac_method

    def set_nac_q_direction(self, nac_q_direction):
        self._nac_q_direction = nac_q_direction

    def get_nac_q_direction(self):
        return self._nac_q_direction

    def set_primitive_matrix(self, primitive_matrix):
        self._primitive_matrix = primitive_matrix

    def get_primitive_matrix(self):
        return self._primitive_matrix

    def set_qpoints(self, qpoints):
        self._qpoints = qpoints

    def get_qpoints(self):
        return self._qpoints

    def set_read_qpoints(self, read_qpoints):
        self._read_qpoints = read_qpoints

    def get_read_qpoints(self):
        return self._read_qpoints

    def set_sigma(self, sigma):
        self._sigma = sigma

    def get_sigma(self):
        return self._sigma

    def set_supercell_matrix(self, matrix):
        self._supercell_matrix = matrix

    def get_supercell_matrix(self):
        return self._supercell_matrix

    def set_symmetry_tolerance(self, symmetry_tolerance):
        self._symmetry_tolerance = symmetry_tolerance

    def get_symmetry_tolerance(self):
        return self._symmetry_tolerance

    def set_temperatures(self, temperatures):
        self._temperatures = temperatures

    def get_temperatures(self):
        return self._temperatures

    def set_temperature_step(self, tstep):
        self._tstep = tstep

    def get_temperature_step(self):
        return self._tstep

    def set_time_reversal_symmetry(self, time_reversal_symmetry=True):
        self._is_time_reversal_symmetry = time_reversal_symmetry

    def get_time_reversal_symmetry(self):
        return self._is_time_reversal_symmetry


# Parse phonopy setting filen
class ConfParser(object):
    def __init__(self, filename=None, args=None):
        self._confs = {}
        self._parameters = {}
        self._args = args
        self._filename = filename

    def get_configures(self):
        return self._confs

    def get_settings(self):
        return self._settings

    def setting_error(self, message):
        print(message)
        print("Please check the setting tags and options.")
        sys.exit(1)

    def read_file(self):
        file = open(self._filename, 'r')
        is_continue = False
        left = None

        for line in file:
            if line.strip() == '':
                is_continue = False
                continue

            if line.strip()[0] == '#':
                is_continue = False
                continue

            if is_continue and left is not None:
                self._confs[left] += line.strip()
                self._confs[left] = self._confs[left].replace('+++', ' ')
                is_continue = False

            if line.find('=') != -1:
                left, right = [x.strip() for x in line.split('=')]
                self._confs[left.lower()] = right

            if line.find('+++') != -1:
                is_continue = True

    def read_options(self):
        arg_list = vars(self._args)
        if 'band_indices' in arg_list:
            band_indices = self._args.band_indices
            if band_indices is not None:
                if type(band_indices) is list:
                    self._confs['band_indices'] = " ".join(band_indices)
                else:
                    self._confs['band_indices'] = band_indices

        if 'band_paths' in arg_list:
            if self._args.band_paths is not None:
                if type(self._args.band_paths) is list:
                    self._confs['band'] = " ".join(self._args.band_paths)
                else:
                    self._confs['band'] = self._args.band_paths

        if 'band_points' in arg_list:
            if self._args.band_points is not None:
                self._confs['band_points'] = self._args.band_points

        if 'cell_filename' in arg_list:
            if self._args.cell_filename is not None:
                self._confs['cell_filename'] = self._args.cell_filename

        if 'cutoff_frequency' in arg_list:
            if self._args.cutoff_frequency:
                self._confs['cutoff_frequency'] = self._args.cutoff_frequency

        if 'displacement_distance' in arg_list:
            if self._args.displacement_distance:
                self._confs['displacement_distance'] = \
                    self._args.displacement_distance

        if 'dynamical_matrix_decimals' in arg_list:
            if self._args.dynamical_matrix_decimals:
                self._confs['dm_decimals'] = \
                    self._args.dynamical_matrix_decimals

        if 'fc_calculator' in arg_list:
            if self._args.fc_calculator:
                self._confs['fc_calculator'] = self._args.fc_calculator

        if 'fc_calculator_options' in arg_list:
            fc_calc_opt = self._args.fc_calculator_options
            if fc_calc_opt:
                self._confs['fc_calculator_options'] = fc_calc_opt

        if 'fc_symmetry' in arg_list:
            if self._args.fc_symmetry:
                self._confs['fc_symmetry'] = '.true.'

        if 'force_constants_decimals' in arg_list:
            if self._args.force_constants_decimals:
                self._confs['fc_decimals'] = \
                    self._args.force_constants_decimals

        if 'fpitch' in arg_list:
            if self._args.fpitch:
                self._confs['fpitch'] = self._args.fpitch

        if 'frequency_conversion_factor' in arg_list:
            freq_factor = self._args.frequency_conversion_factor
            if freq_factor:
                self._confs['frequency_conversion_factor'] = freq_factor

        if 'frequency_scale_factor' in self._args:
            freq_scale = self._args.frequency_scale_factor
            if freq_scale is not None:
                self._confs['frequency_scale_factor'] = freq_scale

        if 'gv_delta_q' in arg_list:
            if self._args.gv_delta_q:
                self._confs['gv_delta_q'] = self._args.gv_delta_q

        if 'is_band_const_interval' in arg_list:
            if self._args.is_band_const_interval:
                self._confs['band_const_interval'] = '.true.'

        if 'is_eigenvectors' in arg_list:
            if self._args.is_eigenvectors:
                self._confs['eigenvectors'] = '.true.'

        if 'is_nac' in arg_list:
            if self._args.is_nac:
                self._confs['nac'] = '.true.'

        if 'is_nodiag' in arg_list:
            if self._args.is_nodiag:
                self._confs['diag'] = '.false.'

        if 'is_nomeshsym' in arg_list:
            if self._args.is_nomeshsym:
                self._confs['mesh_symmetry'] = '.false.'

        if 'is_nosym' in arg_list:
            if self._args.is_nosym:
                self._confs['symmetry'] = '.false.'

        if 'is_plusminus_displacements' in arg_list:
            if self._args.is_plusminus_displacements:
                self._confs['pm'] = '.true.'

        if 'is_tetrahedron_method' in arg_list:
            if not self._args.is_tetrahedron_method:
                self._confs['tetrahedron'] = '.false.'

        if 'is_trigonal_displacements' in arg_list:
            if self._args.is_trigonal_displacements:
                self._confs['trigonal'] = '.true.'

        if 'masses' in arg_list:
            if self._args.masses is not None:
                if type(self._args.masses) is list:
                    self._confs['mass'] = " ".join(self._args.masses)
                else:
                    self._confs['mass'] = self._args.masses

        if 'magmoms' in arg_list:
            if self._args.magmoms is not None:
                if type(self._args.magmoms) is list:
                    self._confs['magmom'] = " ".join(self._args.magmoms)
                else:
                    self._confs['magmom'] = self._args.magmoms

        if 'mesh_numbers' in arg_list:
            mesh = self._args.mesh_numbers
            if mesh is not None:
                if type(mesh) is list:
                    self._confs['mesh_numbers'] = " ".join(mesh)
                else:
                    self._confs['mesh_numbers'] = mesh

        if 'num_frequency_points' in arg_list:
            opt_num_freqs = self._args.num_frequency_points
            if opt_num_freqs:
                self._confs['num_frequency_points'] = opt_num_freqs

        # For backword compatibility
        if 'primitive_axis' in arg_list:
            if self._args.primitive_axis is not None:
                if type(self._args.primitive_axis) is list:
                    primitive_axes = " ".join(self._args.primitive_axis)
                    self._confs['primitive_axes'] = primitive_axes
                else:
                    self._confs['primitive_axes'] = self._args.primitive_axis

        if 'primitive_axes' in arg_list:
            if self._args.primitive_axes:
                if type(self._args.primitive_axes) is list:
                    primitive_axes = " ".join(self._args.primitive_axes)
                    self._confs['primitive_axes'] = primitive_axes
                else:
                    self._confs['primitive_axes'] = self._args.primitive_axes

        if 'supercell_dimension' in arg_list:
            dim = self._args.supercell_dimension
            if dim is not None:
                if type(dim) is list:
                    self._confs['dim'] = " ".join(dim)
                else:
                    self._confs['dim'] = dim

        if 'qpoints' in arg_list:
            if self._args.qpoints is not None:
                if type(self._args.qpoints) is list:
                    self._confs['qpoints'] = " ".join(self._args.qpoints)
                else:
                    self._confs['qpoints'] = self._args.qpoints

        if 'nac_q_direction' in arg_list:
            q_dir = self._args.nac_q_direction
            if q_dir is not None:
                if type(q_dir) is list:
                    self._confs['q_direction'] = " ".join(q_dir)
                else:
                    self._confs['q_direction'] = q_dir

        if 'nac_method' in arg_list:
            if self._args.nac_method is not None:
                self._confs['nac_method'] = self._args.nac_method

        if 'read_qpoints' in arg_list:
            if self._args.read_qpoints:
                self._confs['read_qpoints'] = '.true.'

        if 'sigma' in arg_list:
            if self._args.sigma is not None:
                if type(self._args.sigma) is list:
                    self._confs['sigma'] = " ".join(self._args.sigma)
                else:
                    self._confs['sigma'] = self._args.sigma

        if 'symmetry_tolerance' in arg_list:
            if self._args.symmetry_tolerance:
                symtol = self._args.symmetry_tolerance
                self._confs['symmetry_tolerance'] = symtol

        if 'temperature' in self._args:
            if self._args.temperature is not None:
                self._confs['temperature'] = self._args.temperature

        if 'temperatures' in self._args:
            if self._args.temperatures is not None:
                self._confs['temperatures'] = " ".join(self._args.temperatures)

        if 'tmax' in arg_list:
            if self._args.tmax:
                self._confs['tmax'] = self._args.tmax

        if 'tmin' in arg_list:
            if self._args.tmin:
                self._confs['tmin'] = self._args.tmin

        if 'tstep' in arg_list:
            if self._args.tstep:
                self._confs['tstep'] = self._args.tstep

        if 'use_alm' in arg_list:
            if self._args.use_alm:
                self._confs['fc_calculator'] = 'alm'

        if 'use_hiphive' in arg_list:
            if self._args.use_hiphive:
                self._confs['fc_calculator'] = 'hiphive'

    def parse_conf(self):
        confs = self._confs

        for conf_key in confs.keys():
            if conf_key == 'band_indices':
                vals = []
                for sum_set in confs['band_indices'].split(','):
                    vals.append([int(x) - 1 for x in sum_set.split()])
                self.set_parameter('band_indices', vals)

            if conf_key == 'cell_filename':
                self.set_parameter('cell_filename', confs['cell_filename'])

            if conf_key == 'dim':
                matrix = [int(x) for x in confs['dim'].split()]
                if len(matrix) == 9:
                    matrix = np.array(matrix).reshape(3, 3)
                elif len(matrix) == 3:
                    matrix = np.diag(matrix)
                else:
                    self.setting_error(
                        "Number of elements of DIM tag has to be 3 or 9.")

                if matrix.shape == (3, 3):
                    if np.linalg.det(matrix) < 1:
                        self.setting_error(
                            'Determinant of supercell matrix has to be '
                            'positive.')
                    else:
                        self.set_parameter('supercell_matrix', matrix)

            if conf_key in ('primitive_axis', 'primitive_axes'):
                if confs[conf_key].strip().lower() == 'auto':
                    self.set_parameter('primitive_axes', 'auto')
                elif not len(confs[conf_key].split()) == 9:
                    self.setting_error(
                        "Number of elements in %s has to be 9." %
                        conf_key.upper())
                else:
                    p_axis = []
                    for x in confs[conf_key].split():
                        p_axis.append(fracval(x))
                    p_axis = np.array(p_axis).reshape(3, 3)
                    if np.linalg.det(p_axis) < 1e-8:
                        self.setting_error(
                            "%s has to have positive determinant." %
                            conf_key.upper())
                    self.set_parameter('primitive_axes', p_axis)

            if conf_key == 'mass':
                self.set_parameter(
                    'mass',
                    [float(x) for x in confs['mass'].split()])

            if conf_key == 'magmom':
                self.set_parameter(
                    'magmom',
                    [float(x) for x in confs['magmom'].split()])

            if conf_key == 'atom_name':
                self.set_parameter(
                    'atom_name',
                    [x.capitalize() for x in confs['atom_name'].split()])

            if conf_key == 'displacement_distance':
                self.set_parameter('displacement_distance',
                                   float(confs['displacement_distance']))

            if conf_key == 'diag':
                if confs['diag'].lower() == '.false.':
                    self.set_parameter('diag', False)
                elif confs['diag'].lower() == '.true.':
                    self.set_parameter('diag', True)

            if conf_key == 'pm':
                if confs['pm'].lower() == '.false.':
                    self.set_parameter('pm_displacement', False)
                elif confs['pm'].lower() == '.true.':
                    self.set_parameter('pm_displacement', True)

            if conf_key == 'trigonal':
                if confs['trigonal'].lower() == '.false.':
                    self.set_parameter('is_trigonal_displacement', False)
                elif confs['trigonal'].lower() == '.true.':
                    self.set_parameter('is_trigonal_displacement', True)

            if conf_key == 'eigenvectors':
                if confs['eigenvectors'].lower() == '.false.':
                    self.set_parameter('is_eigenvectors', False)
                elif confs['eigenvectors'].lower() == '.true.':
                    self.set_parameter('is_eigenvectors', True)

            if conf_key == 'nac':
                if confs['nac'].lower() == '.false.':
                    self.set_parameter('is_nac', False)
                elif confs['nac'].lower() == '.true.':
                    self.set_parameter('is_nac', True)

            if conf_key == 'symmetry':
                if confs['symmetry'].lower() == '.false.':
                    self.set_parameter('is_symmetry', False)
                    self.set_parameter('is_mesh_symmetry', False)
                elif confs['symmetry'].lower() == '.true.':
                    self.set_parameter('is_symmetry', True)

            if conf_key == 'mesh_symmetry':
                if confs['mesh_symmetry'].lower() == '.false.':
                    self.set_parameter('is_mesh_symmetry', False)
                elif confs['mesh_symmetry'].lower() == '.true.':
                    self.set_parameter('is_mesh_symmetry', True)

            if conf_key == 'rotational':
                if confs['rotational'].lower() == '.false.':
                    self.set_parameter('is_rotational', False)
                elif confs['rotational'].lower() == '.true.':
                    self.set_parameter('is_rotational', True)

            if conf_key == 'fc_calculator':
                self.set_parameter('fc_calculator', confs['fc_calculator'])

            if conf_key == 'fc_calculator_options':
                self.set_parameter('fc_calculator_options',
                                   confs['fc_calculator_options'])

            if conf_key == 'fc_symmetry':
                if confs['fc_symmetry'].lower() == '.false.':
                    self.set_parameter('fc_symmetry', False)
                elif confs['fc_symmetry'].lower() == '.true.':
                    self.set_parameter('fc_symmetry', True)
                else:
                    self.setting_error(
                        "FC_SYMMETRY has to be specified by .TRUE. or .FALSE.")

            if conf_key == 'fc_decimals':
                self.set_parameter('fc_decimals', confs['fc_decimals'])

            if conf_key == 'dm_decimals':
                self.set_parameter('dm_decimals', confs['dm_decimals'])

            if conf_key in ['mesh_numbers', 'mp', 'mesh']:
                vals = [x for x in confs[conf_key].split()]
                if len(vals) == 1:
                    self.set_parameter('mesh_numbers', float(vals[0]))
                elif len(vals) < 3:
                    self.setting_error("Mesh numbers are incorrectly set.")
                else:
                    self.set_parameter('mesh_numbers',
                                       [int(x) for x in vals[:3]])

            if conf_key == 'band_points':
                self.set_parameter('band_points', int(confs['band_points']))

            if conf_key == 'band_const_interval':
                if confs['band_const_interval'].lower() == '.false.':
                    self.set_parameter('is_band_const_interval', False)
                elif confs['band_const_interval'].lower() == '.true.':
                    self.set_parameter('is_band_const_interval', True)

            if conf_key == 'band':
                bands = []
                if confs['band'].strip().lower() == 'auto':
                    self.set_parameter('band_paths', 'auto')
                else:
                    for section in confs['band'].split(','):
                        points = [fracval(x) for x in section.split()]
                        if len(points) % 3 != 0 or len(points) < 6:
                            self.setting_error("BAND is incorrectly set.")
                            break
                        bands.append(np.array(points).reshape(-1, 3))
                    self.set_parameter('band_paths', bands)

            if conf_key == 'qpoints':
                if confs['qpoints'].lower() == '.true.':
                    self.set_parameter('read_qpoints', True)
                elif confs['qpoints'].lower() == '.false.':
                    self.set_parameter('read_qpoints', False)
                else:
                    vals = [fracval(x) for x in confs['qpoints'].split()]
                    if len(vals) == 0 or len(vals) % 3 != 0:
                        self.setting_error("Q-points are incorrectly set.")
                    else:
                        self.set_parameter('qpoints',
                                           list(np.reshape(vals, (-1, 3))))

            if conf_key == 'read_qpoints':
                if confs['read_qpoints'].lower() == '.false.':
                    self.set_parameter('read_qpoints', False)
                elif confs['read_qpoints'].lower() == '.true.':
                    self.set_parameter('read_qpoints', True)

            if conf_key == 'nac_method':
                self.set_parameter('nac_method', confs['nac_method'].lower())

            if conf_key == 'q_direction':
                q_direction = [fracval(x)
                               for x in confs['q_direction'].split()]
                if len(q_direction) < 3:
                    self.setting_error("Number of elements of q_direction "
                                       "is less than 3")
                else:
                    self.set_parameter('nac_q_direction', q_direction)

            if conf_key == 'frequency_conversion_factor':
                val = float(confs['frequency_conversion_factor'])
                self.set_parameter('frequency_conversion_factor', val)

            if conf_key == 'frequency_scale_factor':
                self.set_parameter('frequency_scale_factor',
                                   float(confs['frequency_scale_factor']))

            if conf_key == 'fpitch':
                val = float(confs['fpitch'])
                self.set_parameter('fpitch', val)

            if conf_key == 'num_frequency_points':
                val = int(confs['num_frequency_points'])
                self.set_parameter('num_frequency_points', val)

            if conf_key == 'cutoff_frequency':
                val = float(confs['cutoff_frequency'])
                self.set_parameter('cutoff_frequency', val)

            if conf_key == 'sigma':
                vals = [float(x) for x in str(confs['sigma']).split()]
                if len(vals) == 1:
                    self.set_parameter('sigma', vals[0])
                else:
                    self.set_parameter('sigma', vals)

            if conf_key == 'tetrahedron':
                if confs['tetrahedron'].lower() == '.false.':
                    self.set_parameter('is_tetrahedron_method', False)
                if confs['tetrahedron'].lower() == '.true.':
                    self.set_parameter('is_tetrahedron_method', True)

            if conf_key == 'symmetry_tolerance':
                val = float(confs['symmetry_tolerance'])
                self.set_parameter('symmetry_tolerance', val)

            # For multiple T values.
            if conf_key == 'temperatures':
                vals = [fracval(x) for x in confs['temperatures'].split()]
                if len(vals) < 1:
                    self.setting_error("Temperatures are incorrectly set.")
                else:
                    self.set_parameter('temperatures', vals)

            # For single T value.
            if conf_key == 'temperature':
                self.set_parameter('temperatures', [confs['temperature'], ])

            if conf_key == 'tmin':
                val = float(confs['tmin'])
                self.set_parameter('tmin', val)

            if conf_key == 'tmax':
                val = float(confs['tmax'])
                self.set_parameter('tmax', val)

            if conf_key == 'tstep':
                val = float(confs['tstep'])
                self.set_parameter('tstep', val)

            # Group velocity finite difference
            if conf_key == 'gv_delta_q':
                self.set_parameter('gv_delta_q', float(confs['gv_delta_q']))

            # Use ALM for generating force constants
            if conf_key == 'alm':
                if confs['alm'].lower() == '.true.':
                    self.set_parameter('alm', True)

    def set_parameter(self, key, val):
        self._parameters[key] = val

    def set_settings(self):
        params = self._parameters

        # Chemical symbols
        if 'atom_name' in params:
            self._settings.set_chemical_symbols(params['atom_name'])

        # Sets of band indices that are summed
        if 'band_indices' in params:
            self._settings.set_band_indices(params['band_indices'])

        # Filename of input unit cell
        if 'cell_filename' in params:
            self._settings.set_cell_filename(params['cell_filename'])

        # Cutoff frequency
        if 'cutoff_frequency' in params:
            self._settings.set_cutoff_frequency(params['cutoff_frequency'])

        # Diagonal displacement
        if 'diag' in params:
            self._settings.set_is_diagonal_displacement(params['diag'])

        # Distance of finite displacements introduced
        if 'displacement_distance' in params:
            self._settings.set_displacement_distance(
                params['displacement_distance'])

        # Decimals of values of dynamical matrxi
        if 'dm_decimals' in params:
            self._settings.set_dm_decimals(int(params['dm_decimals']))

        # Force constants calculator
        if 'fc_calculator' in params:
            self._settings.set_fc_calculator(params['fc_calculator'])

        # Force constants calculator options as str
        if 'fc_calculator_options' in params:
            self._settings.set_fc_calculator_options(
                params['fc_calculator_options'])

        # Decimals of values of force constants
        if 'fc_decimals' in params:
            self._settings.set_fc_decimals(int(params['fc_decimals']))

        # Enforce translational invariance and index permutation symmetry
        # to force constants?
        if 'fc_symmetry' in params:
            self._settings.set_fc_symmetry(params['fc_symmetry'])

        # Frequency unit conversion factor
        if 'frequency_conversion_factor' in params:
            self._settings.set_frequency_conversion_factor(
                params['frequency_conversion_factor'])

        # This scale factor is multiplied to force constants by
        # fc * scale_factor ** 2, therefore only changes
        # frequencies but does not change NAC part.
        if 'frequency_scale_factor' in params:
            self._settings.set_frequency_scale_factor(
                params['frequency_scale_factor'])

        # Spectram drawing step
        if 'fpitch' in params:
            self._settings.set_frequency_pitch(params['fpitch'])

        # Number of sampling points for spectram drawing
        if 'num_frequency_points' in params:
            self._settings.set_num_frequency_points(params['num_frequency_points'])

        # Group velocity finite difference
        if 'gv_delta_q' in params:
            self._settings.set_group_velocity_delta_q(params['gv_delta_q'])

        # Mesh sampling numbers
        if 'mesh_numbers' in params:
            self._settings.set_mesh_numbers(params['mesh_numbers'])

        # Is getting eigenvectors?
        if 'is_eigenvectors' in params:
            self._settings.set_is_eigenvectors(params['is_eigenvectors'])

        # Is reciprocal mesh symmetry searched?
        if 'is_mesh_symmetry' in params:
            self._settings.set_is_mesh_symmetry(params['is_mesh_symmetry'])

        # Non analytical term correction?
        if 'is_nac' in params:
            self._settings.set_is_nac(params['is_nac'])

        # Is rotational invariance ?
        if 'is_rotational' in params:
            self._settings.set_is_rotational_invariance(params['is_rotational'])

        # Is crystal symmetry searched?
        if 'is_symmetry' in params:
            self._settings.set_is_symmetry(params['is_symmetry'])

        # Tetrahedron method
        if 'is_tetrahedron_method' in params:
            self._settings.set_is_tetrahedron_method(
                params['is_tetrahedron_method'])

        # Trigonal displacement
        if 'is_trigonal_displacement' in params:
            self._settings.set_is_trigonal_displacement(
                params['is_trigonal_displacement'])

        # Magnetic moments
        if 'magmom' in params:
            self._settings.set_magnetic_moments(params['magmom'])

        # Atomic mass
        if 'mass' in params:
            self._settings.set_masses(params['mass'])

        # Plus minus displacement
        if 'pm_displacement' in params:
            self._settings.set_is_plusminus_displacement(
                params['pm_displacement'])

        # Primitive cell shape
        if 'primitive_axes' in params:
            self._settings.set_primitive_matrix(params['primitive_axes'])

        # Q-points mode
        if 'qpoints' in params:
            self._settings.set_qpoints(params['qpoints'])

        if 'read_qpoints' in params:
            if params['read_qpoints']:
                self._settings.set_read_qpoints(params['read_qpoints'])

        # non analytical term correction method
        if 'nac_method' in params:
            self._settings.set_nac_method(params['nac_method'])

        # q-direction for non analytical term correction
        if 'nac_q_direction' in params:
            self._settings.set_nac_q_direction(params['nac_q_direction'])

        # Smearing width
        if 'sigma' in params:
            self._settings.set_sigma(params['sigma'])

        # Symmetry tolerance
        if 'symmetry_tolerance' in params:
            self._settings.set_symmetry_tolerance(params['symmetry_tolerance'])

        # Supercell size
        if 'supercell_matrix' in params:
            self._settings.set_supercell_matrix(params['supercell_matrix'])

        # Temperatures or temerature range
        if 'temperatures' in params:
            self._settings.set_temperatures(params['temperatures'])
        if 'tmax' in params:
            self._settings.set_max_temperature(params['tmax'])
        if 'tmin' in params:
            self._settings.set_min_temperature(params['tmin'])
        if 'tstep' in params:
            self._settings.set_temperature_step(params['tstep'])

        # Band paths
        # BAND = 0.0 0.0 0.0  0.5 0.0 0.0  0.5 0.5 0.0  0.0 0.0 0.0  0.5 0.5 0.5
        # [array([[ 0. ,  0. ,  0. ],
        #         [ 0.5,  0. ,  0. ],
        #         [ 0.5,  0.5,  0. ],
        #         [ 0. ,  0. ,  0. ],
        #         [ 0.5,  0.5,  0.5]])]
        #
        # BAND = 0.0 0.0 0.0  0.5 0.0 0.0, 0.5 0.5 0.0  0.0 0.0 0.0  0.5 0.5 0.5
        # [array([[ 0. ,  0. ,  0. ],
        #         [ 0.5,  0. ,  0. ]]),
        #  array([[ 0.5,  0.5,  0. ],
        #         [ 0. ,  0. ,  0. ],
        #         [ 0.5,  0.5,  0.5]])]
        # or
        # BAND = AUTO
        if 'band_paths' in params:
            self._settings.set_band_paths(params['band_paths'])

        # This number includes end points
        if 'band_points' in params:
            self._settings.set_band_points(params['band_points'])

        if 'is_band_const_interval' in params:
            self._settings.set_is_band_const_interval(
                params['is_band_const_interval'])

        # Use ALM to generating force constants
        if 'alm' in params:
            self._settings.set_use_alm(params['alm'])


#
# For phonopy
#
class PhonopySettings(Settings):
    def __init__(self):
        Settings.__init__(self)

        self._anime_band_index = None
        self._anime_amplitude = None
        self._anime_division = None
        self._anime_qpoint = None
        self._anime_shift = None
        self._anime_type = 'v_sim'
        self._band_format = 'yaml'
        self._band_labels = None
        self._band_connection = False
        self._cutoff_radius = None
        self._dos = None
        self._fc_spg_symmetry = False
        self._fits_Debye_model = False
        self._fmax = None
        self._fmin = None
        self._irreps_q_point = None
        self._irreps_tolerance = 1e-5
        self._is_dos_mode = False
        self._is_full_fc = False
        self._is_group_velocity = False
        self._is_gamma_center = False
        self._is_hdf5 = False
        self._is_legacy_plot = False
        self._is_little_cogroup = False
        self._is_moment = False
        self._is_plusminus_displacement = 'auto'
        self._is_thermal_displacements = False
        self._is_thermal_displacement_matrices = False
        self._is_thermal_distances = False
        self._is_thermal_properties = False
        self._is_projected_thermal_properties = False
        self._lapack_solver = False
        self._mesh_shift = None
        self._mesh_format = 'yaml'
        self._modulation = None
        self._moment_order = None
        self._random_displacements = None
        self._pdos_indices = None
        self._pretend_real = False
        self._projection_direction = None
        self._random_seed = None
        self._qpoints_format = 'yaml'
        self._read_force_constants = False
        self._readfc_format = 'text'
        self._run_mode = None
        self._show_irreps = False
        self._thermal_atom_pairs = None
        self._thermal_displacement_matrix_temperatue = None
        self._write_dynamical_matrices = False
        self._write_mesh = True
        self._write_force_constants = False
        self._writefc_format = 'text'
        self._xyz_projection = False

    def set_anime_band_index(self, band_index):
        self._anime_band_index = band_index

    def get_anime_band_index(self):
        return self._anime_band_index

    def set_anime_amplitude(self, amplitude):
        self._anime_amplitude = amplitude

    def get_anime_amplitude(self):
        return self._anime_amplitude

    def set_anime_division(self, division):
        self._anime_division = division

    def get_anime_division(self):
        return self._anime_division

    def set_anime_qpoint(self, qpoint):
        self._anime_qpoint = qpoint

    def get_anime_qpoint(self):
        return self._anime_qpoint

    def set_anime_shift(self, shift):
        self._anime_shift = shift

    def get_anime_shift(self):
        return self._anime_shift

    def set_anime_type(self, anime_type):
        self._anime_type = anime_type

    def get_anime_type(self):
        return self._anime_type

    def set_band_format(self, band_format):
        self._band_format = band_format

    def get_band_format(self):
        return self._band_format

    def set_band_labels(self, labels):
        self._band_labels = labels

    def get_band_labels(self):
        return self._band_labels

    def set_cutoff_radius(self, cutoff_radius):
        self._cutoff_radius = cutoff_radius

    def get_cutoff_radius(self):
        return self._cutoff_radius

    def set_fc_spg_symmetry(self, fc_spg_symmetry):
        self._fc_spg_symmetry = fc_spg_symmetry

    def get_fc_spg_symmetry(self):
        return self._fc_spg_symmetry

    def set_fits_Debye_model(self, fits_Debye_model):
        self._fits_Debye_model = fits_Debye_model

    def get_fits_Debye_model(self):
        return self._fits_Debye_model

    def set_max_frequency(self, fmax):
        self._fmax = fmax

    def get_max_frequency(self):
        return self._fmax

    def set_mesh_shift(self, mesh_shift):
        self._mesh_shift = mesh_shift

    def get_mesh_shift(self):
        return self._mesh_shift

    def set_min_frequency(self, fmin):
        self._fmin = fmin

    def get_min_frequency(self):
        return self._fmin

    def set_irreps_q_point(self, q_point):
        self._irreps_q_point = q_point

    def get_irreps_q_point(self):
        return self._irreps_q_point

    def set_irreps_tolerance(self, tolerance):
        self._irreps_tolerance = tolerance

    def get_irreps_tolerance(self):
        return self._irreps_tolerance

    def set_is_band_connection(self, band_connection):
        self._band_connection = band_connection

    def get_is_band_connection(self):
        return self._band_connection

    def set_is_dos_mode(self, is_dos_mode):
        self._is_dos_mode = is_dos_mode

    def get_is_dos_mode(self):
        return self._is_dos_mode

    def set_is_full_fc(self, is_full_fc):
        self._is_full_fc = is_full_fc

    def get_is_full_fc(self):
        return self._is_full_fc

    def set_is_gamma_center(self, is_gamma_center):
        self._is_gamma_center = is_gamma_center

    def get_is_gamma_center(self):
        return self._is_gamma_center

    def set_is_group_velocity(self, is_group_velocity):
        self._is_group_velocity = is_group_velocity

    def get_is_group_velocity(self):
        return self._is_group_velocity

    def set_is_hdf5(self, is_hdf5):
        self._is_hdf5 = is_hdf5

    def get_is_hdf5(self):
        return self._is_hdf5

    def set_is_legacy_plot(self, is_legacy_plot):
        self._is_legacy_plot = is_legacy_plot

    def get_is_legacy_plot(self):
        return self._is_legacy_plot

    def set_is_little_cogroup(self, is_little_cogroup):
        self._is_little_cogroup = is_little_cogroup

    def get_is_little_cogroup(self):
        return self._is_little_cogroup

    def set_is_moment(self, is_moment):
        self._is_moment = is_moment

    def get_is_moment(self):
        return self._is_moment

    def set_is_projected_thermal_properties(self, is_ptp):
        self._is_projected_thermal_properties = is_ptp

    def get_is_projected_thermal_properties(self):
        return self._is_projected_thermal_properties

    def set_is_thermal_displacements(self, is_thermal_displacements):
        self._is_thermal_displacements = is_thermal_displacements

    def get_is_thermal_displacements(self):
        return self._is_thermal_displacements

    def set_is_thermal_displacement_matrices(self, is_displacement_matrices):
        self._is_thermal_displacement_matrices = is_displacement_matrices

    def get_is_thermal_displacement_matrices(self):
        return self._is_thermal_displacement_matrices

    def set_is_thermal_distances(self, is_thermal_distances):
        self._is_thermal_distances = is_thermal_distances

    def get_is_thermal_distances(self):
        return self._is_thermal_distances

    def set_is_thermal_properties(self, is_thermal_properties):
        self._is_thermal_properties = is_thermal_properties

    def get_is_thermal_properties(self):
        return self._is_thermal_properties

    def set_lapack_solver(self, lapack_solver):
        self._lapack_solver = lapack_solver

    def get_lapack_solver(self):
        return self._lapack_solver

    def set_mesh_format(self, mesh_format):
        self._mesh_format = mesh_format

    def get_mesh_format(self):
        return self._mesh_format

    def set_modulation(self, modulation):
        self._modulation = modulation

    def get_modulation(self):
        return self._modulation

    def set_moment_order(self, moment_order):
        self._moment_order = moment_order

    def get_moment_order(self):
        return self._moment_order

    def set_random_displacements(self, random_displacements):
        self._random_displacements = random_displacements

    def get_random_displacements(self):
        return self._random_displacements

    def set_pdos_indices(self, indices):
        self._pdos_indices = indices

    def get_pdos_indices(self):
        return self._pdos_indices

    def set_pretend_real(self, pretend_real):
        self._pretend_real = pretend_real

    def get_pretend_real(self):
        return self._pretend_real

    def set_projection_direction(self, direction):
        self._projection_direction = direction

    def get_projection_direction(self):
        return self._projection_direction

    def set_qpoints_format(self, qpoints_format):
        self._qpoints_format = qpoints_format

    def get_qpoints_format(self):
        return self._qpoints_format

    def set_random_seed(self, random_seed):
        self._random_seed = random_seed

    def get_random_seed(self):
        return self._random_seed

    def set_read_force_constants(self, read_force_constants):
        self._read_force_constants = read_force_constants

    def get_read_force_constants(self):
        return self._read_force_constants

    def set_readfc_format(self, readfc_format):
        self._readfc_format = readfc_format

    def get_readfc_format(self):
        return self._readfc_format

    def set_run_mode(self, run_mode):
        modes = ['qpoints',
                 'mesh',
                 'band',
                 'band_mesh',
                 'anime',
                 'modulation',
                 'displacements',
                 'irreps']
        for mode in modes:
            if run_mode.lower() == mode:
                self._run_mode = run_mode

    def get_run_mode(self):
        return self._run_mode

    def set_thermal_property_range(self, tmin, tmax, tstep):
        self._tmax = tmax
        self._tmin = tmin
        self._tstep = tstep

    def get_thermal_property_range(self):
        return {'min':  self._tmin,
                'max':  self._tmax,
                'step': self._tstep}

    def set_thermal_atom_pairs(self, atom_pairs):
        self._thermal_atom_pairs = atom_pairs

    def get_thermal_atom_pairs(self):
        return self._thermal_atom_pairs

    def set_thermal_displacement_matrix_temperature(self, t):
        self._thermal_displacement_matrix_temperatue = t

    def get_thermal_displacement_matrix_temperature(self):
        return self._thermal_displacement_matrix_temperatue

    def set_show_irreps(self, show_irreps):
        self._show_irreps = show_irreps

    def get_show_irreps(self):
        return self._show_irreps

    def set_write_dynamical_matrices(self, write_dynamical_matrices):
        self._write_dynamical_matrices = write_dynamical_matrices

    def get_write_dynamical_matrices(self):
        return self._write_dynamical_matrices

    def set_write_force_constants(self, write_force_constants):
        self._write_force_constants = write_force_constants

    def get_write_force_constants(self):
        return self._write_force_constants

    def set_write_mesh(self, write_mesh):
        self._write_mesh = write_mesh

    def get_write_mesh(self):
        return self._write_mesh

    def set_writefc_format(self, writefc_format):
        self._writefc_format = writefc_format

    def get_writefc_format(self):
        return self._writefc_format

    def set_xyz_projection(self, xyz_projection):
        self._xyz_projection = xyz_projection

    def get_xyz_projection(self):
        return self._xyz_projection


class PhonopyConfParser(ConfParser):
    def __init__(self, filename=None, args=None):
        # This is fragile implementation.
        # Remember that options have to be activated only
        # when it changes the default value, i.e.,
        # _read_options has to be written in this way.

        self._settings = PhonopySettings()
        confs = {}
        if filename is not None:
            ConfParser.__init__(self, filename=filename)
            self.read_file()  # store .conf file setting in self._confs
            self._parse_conf()  # self.parameters[key] = val
            self._set_settings()  # self.parameters -> PhonopySettings
            confs.update(self._confs)
        if args is not None:
            # To invoke ConfParser.__init__() to flush variables.
            ConfParser.__init__(self, args=args)
            self._read_options()  # store options in self._confs
            self._parse_conf()  # self.parameters[key] = val
            self._set_settings()  # self.parameters -> PhonopySettings
            confs.update(self._confs)
        self._confs = confs

    def _read_options(self):
        self.read_options()  # store data in self._confs
        arg_list = vars(self._args)
        if 'band_format' in arg_list:
            if self._args.band_format:
                self._confs['band_format'] = self._args.band_format

        if 'band_labels' in arg_list:
            if self._args.band_labels is not None:
                self._confs['band_labels'] = " ".join(self._args.band_labels)

        if 'is_displacement' in arg_list:
            if self._args.is_displacement:
                self._confs['create_displacements'] = '.true.'

        if 'is_gamma_center' in arg_list:
            if self._args.is_gamma_center:
                self._confs['gamma_center'] = '.true.'

        if 'is_dos_mode' in arg_list:
            if self._args.is_dos_mode:
                self._confs['dos'] = '.true.'

        if 'pdos' in arg_list:
            if self._args.pdos is not None:
                self._confs['pdos'] = " ".join(self._args.pdos)

        if 'xyz_projection' in arg_list:
            if self._args.xyz_projection:
                self._confs['xyz_projection'] = '.true.'

        if 'fc_spg_symmetry' in arg_list:
            if self._args.fc_spg_symmetry:
                self._confs['fc_spg_symmetry'] = '.true.'

        if 'is_full_fc' in arg_list:
            if self._args.is_full_fc:
                self._confs['full_force_constants'] = '.true.'

        if 'fits_debye_model' in arg_list:
            if self._args.fits_debye_model:
                self._confs['debye_model'] = '.true.'

        if 'fmax' in arg_list:
            if self._args.fmax:
                self._confs['fmax'] = self._args.fmax

        if 'fmin' in arg_list:
            if self._args.fmin:
                self._confs['fmin'] = self._args.fmin

        if 'is_thermal_properties' in arg_list:
            if self._args.is_thermal_properties:
                self._confs['tprop'] = '.true.'

        if 'pretend_real' in arg_list:
            if self._args.pretend_real:
                self._confs['pretend_real'] = '.true.'

        if 'is_projected_thermal_properties' in arg_list:
            if self._args.is_projected_thermal_properties:
                self._confs['ptprop'] = '.true.'

        if 'is_thermal_displacements' in arg_list:
            if self._args.is_thermal_displacements:
                self._confs['tdisp'] = '.true.'

        if 'is_thermal_displacement_matrices' in arg_list:
            if self._args.is_thermal_displacement_matrices:
                self._confs['tdispmat'] = '.true.'

        if 'thermal_displacement_matrices_cif' in arg_list:
            opt_tdm_cif = self._args.thermal_displacement_matrices_cif
            if opt_tdm_cif:
                self._confs['tdispmat_cif'] = opt_tdm_cif

        if 'projection_direction' in arg_list:
            opt_proj_dir = self._args.projection_direction
            if opt_proj_dir is not None:
                self._confs['projection_direction'] = " ".join(opt_proj_dir)

        if 'read_force_constants' in arg_list:
            if self._args.read_force_constants:
                self._confs['read_force_constants'] = '.true.'

        if 'write_force_constants' in arg_list:
            if self._args.write_force_constants:
                self._confs['write_force_constants'] = '.true.'

        if 'readfc_format' in arg_list:
            if self._args.readfc_format:
                self._confs['readfc_format'] = self._args.readfc_format

        if 'writefc_format' in arg_list:
            if self._args.writefc_format:
                self._confs['writefc_format'] = self._args.writefc_format

        if 'fc_format' in arg_list:
            if self._args.fc_format:
                self._confs['fc_format'] = self._args.fc_format

        if 'is_hdf5' in arg_list:
            if self._args.is_hdf5:
                self._confs['hdf5'] = '.true.'

        if 'write_dynamical_matrices' in arg_list:
            if self._args.write_dynamical_matrices:
                self._confs['writedm'] = '.true.'

        if 'write_mesh' in arg_list:
            if not self._args.write_mesh:
                self._confs['write_mesh'] = '.false.'

        if 'mesh_format' in arg_list:
            if self._args.mesh_format:
                self._confs['mesh_format'] = self._args.mesh_format

        if 'qpoints_format' in arg_list:
            if self._args.qpoints_format:
                self._confs['qpoints_format'] = self._args.qpoints_format

        if 'irreps_qpoint' in arg_list:
            if self._args.irreps_qpoint is not None:
                self._confs['irreps'] = " ".join(self._args.irreps_qpoint)

        if 'show_irreps' in arg_list:
            if self._args.show_irreps:
                self._confs['show_irreps'] = '.true.'

        if 'is_little_cogroup' in arg_list:
            if self._args.is_little_cogroup:
                self._confs['little_cogroup'] = '.true.'

        if 'is_legacy_plot' in arg_list:
            if self._args.is_legacy_plot:
                self._confs['legacy_plot'] = '.true.'

        if 'is_band_connection' in arg_list:
            if self._args.is_band_connection:
                self._confs['band_connection'] = '.true.'

        if 'cutoff_radius' in arg_list:
            if self._args.cutoff_radius:
                self._confs['cutoff_radius'] = self._args.cutoff_radius

        if 'modulation' in arg_list:
            if self._args.modulation:
                self._confs['modulation'] = " ".join(self._args.modulation)

        if 'anime' in arg_list:
            if self._args.anime:
                self._confs['anime'] = " ".join(self._args.anime)

        if 'is_group_velocity' in arg_list:
            if self._args.is_group_velocity:
                self._confs['group_velocity'] = '.true.'

        if 'is_moment' in arg_list:
            if self._args.is_moment:
                self._confs['moment'] = '.true.'

        if 'moment_order' in arg_list:
            if self._args.moment_order:
                self._confs['moment_order'] = self._args.moment_order

        if 'random_displacements' in arg_list:
            nrand = self._args.random_displacements
            if nrand:
                self._confs['random_displacements'] = nrand

        if 'random_seed' in arg_list:
            if self._args.random_seed:
                seed = self._args.random_seed
                if (np.issubdtype(type(seed), np.integer) and
                    seed >= 0 and seed < 2 ** 32):
                    self._confs['random_seed'] = seed

        # Overwrite
        if 'is_check_symmetry' in arg_list:
            if self._args.is_check_symmetry:
                # Dummy 'dim' setting for sym-check
                self._confs['dim'] = '1 1 1'

        if 'lapack_solver' in arg_list:
            if self._args.lapack_solver:
                self._confs['lapack_solver'] = '.true.'

    def _parse_conf(self):
        self.parse_conf()
        confs = self._confs

        for conf_key in confs.keys():
            if conf_key == 'create_displacements':
                if confs['create_displacements'].lower() == '.true.':
                    self.set_parameter('create_displacements', True)

            if conf_key == 'band_format':
                self.set_parameter('band_format', confs['band_format'].lower())

            if conf_key == 'band_labels':
                labels = [x for x in confs['band_labels'].split()]
                self.set_parameter('band_labels', labels)

            if conf_key == 'band_connection':
                if confs['band_connection'].lower() == '.true.':
                    self.set_parameter('band_connection', True)

            if conf_key == 'legacy_plot':
                if confs['legacy_plot'].lower() == '.true.':
                    self.set_parameter('legacy_plot', True)

            if conf_key == 'force_constants':
                self.set_parameter('force_constants',
                                   confs['force_constants'].lower())

            if conf_key == 'read_force_constants':
                if confs['read_force_constants'].lower() == '.true.':
                    self.set_parameter('read_force_constants', True)

            if conf_key == 'write_force_constants':
                if confs['write_force_constants'].lower() == '.true.':
                    self.set_parameter('write_force_constants', True)

            if conf_key == 'full_force_constants':
                if confs['full_force_constants'].lower() == '.true.':
                    self.set_parameter('is_full_fc', True)

            if conf_key == 'cutoff_radius':
                val = float(confs['cutoff_radius'])
                self.set_parameter('cutoff_radius', val)

            if conf_key == 'writedm':
                if confs['writedm'].lower() == '.true.':
                    self.set_parameter('write_dynamical_matrices', True)

            if conf_key == 'write_mesh':
                if confs['write_mesh'].lower() == '.false.':
                    self.set_parameter('write_mesh', False)

            if conf_key == 'hdf5':
                if confs['hdf5'].lower() == '.true.':
                    self.set_parameter('hdf5', True)

            if conf_key == 'mp_shift':
                vals = [fracval(x) for x in confs['mp_shift'].split()]
                if len(vals) < 3:
                    self.setting_error("MP_SHIFT is incorrectly set.")
                self.set_parameter('mp_shift', vals[:3])

            if conf_key == 'mesh_format':
                self.set_parameter('mesh_format', confs['mesh_format'].lower())

            if conf_key == 'qpoints_format':
                self.set_parameter('qpoints_format',
                                   confs['qpoints_format'].lower())

            if conf_key == 'time_reversal_symmetry':
                if confs['time_reversal_symmetry'].lower() == '.false.':
                    self.set_parameter('is_time_reversal_symmetry', False)

            if conf_key == 'gamma_center':
                if confs['gamma_center'].lower() == '.true.':
                    self.set_parameter('is_gamma_center', True)

            if conf_key == 'fc_spg_symmetry':
                if confs['fc_spg_symmetry'].lower() == '.true.':
                    self.set_parameter('fc_spg_symmetry', True)

            if conf_key == 'readfc_format':
                self.set_parameter('readfc_format',
                                   confs['readfc_format'].lower())

            if conf_key == 'writefc_format':
                self.set_parameter('writefc_format',
                                   confs['writefc_format'].lower())

            if conf_key == 'fc_format':
                self.set_parameter('readfc_format', confs['fc_format'].lower())
                self.set_parameter('writefc_format',
                                   confs['fc_format'].lower())

            # Animation
            if conf_key == 'anime':
                vals = []
                data = confs['anime'].split()
                if len(data) < 3:
                    self.setting_error("ANIME is incorrectly set.")
                else:
                    self.set_parameter('anime', data)

            if conf_key == 'anime_type':
                anime_type = confs['anime_type'].lower()
                if anime_type in ('arc', 'v_sim', 'poscar', 'xyz', 'jmol'):
                    self.set_parameter('anime_type', anime_type)
                else:
                    self.setting_error("%s is not available for ANIME_TYPE tag."
                                       % confs['anime_type'])

            # Modulation
            if conf_key == 'modulation':
                self._parse_conf_modulation(confs['modulation'])

            # Character table
            if conf_key == 'irreps':
                vals = [fracval(x) for x in confs['irreps'].split()]
                if len(vals) == 3 or len(vals) == 4:
                    self.set_parameter('irreps_qpoint', vals)
                else:
                    self.setting_error("IRREPS is incorrectly set.")

            if conf_key == 'show_irreps':
                if confs['show_irreps'].lower() == '.true.':
                    self.set_parameter('show_irreps', True)

            if conf_key == 'little_cogroup':
                if confs['little_cogroup'].lower() == '.true.':
                    self.set_parameter('little_cogroup', True)

            # DOS
            if conf_key == 'pdos':
                if confs['pdos'].strip().lower() == 'auto':
                    self.set_parameter('pdos', 'auto')
                else:
                    vals = []
                    for index_set in confs['pdos'].split(','):
                        vals.append([int(x) - 1 for x in index_set.split()])
                    self.set_parameter('pdos', vals)

            if conf_key == 'xyz_projection':
                if confs['xyz_projection'].lower() == '.true.':
                    self.set_parameter('xyz_projection', True)

            if conf_key == 'dos':
                if confs['dos'].lower() == '.true.':
                    self.set_parameter('dos', True)

            if conf_key == 'debye_model':
                if confs['debye_model'].lower() == '.true.':
                    self.set_parameter('fits_debye_model', True)

            if conf_key == 'dos_range':
                vals = [float(x) for x in confs['dos_range'].split()]
                self.set_parameter('dos_range', vals)

            if conf_key == 'fmax':
                self.set_parameter('fmax', float(confs['fmax']))

            if conf_key == 'fmin':
                self.set_parameter('fmin', float(confs['fmin']))

            # Thermal properties
            if conf_key == 'tprop':
                if confs['tprop'].lower() == '.true.':
                    self.set_parameter('tprop', True)

            # Projected thermal properties
            if conf_key == 'ptprop':
                if confs['ptprop'].lower() == '.true.':
                    self.set_parameter('ptprop', True)

            # Use imaginary frequency as real for thermal property calculation
            if conf_key == 'pretend_real':
                if confs['pretend_real'].lower() == '.true.':
                    self.set_parameter('pretend_real', True)

            # Thermal displacement
            if conf_key == 'tdisp':
                if confs['tdisp'].lower() == '.true.':
                    self.set_parameter('tdisp', True)

            # Thermal displacement matrices
            if conf_key == 'tdispmat':
                if confs['tdispmat'].lower() == '.true.':
                    self.set_parameter('tdispmat', True)

            # Write thermal displacement matrices to cif file,
            # for which the temperature to execute is stored.
            if conf_key == 'tdispmat_cif':
                self.set_parameter('tdispmat_cif', float(confs['tdispmat_cif']))

            # Thermal distance
            if conf_key == 'tdistance':
                atom_pairs = []
                for atoms in confs['tdistance'].split(','):
                    pair = [int(x) - 1 for x in atoms.split()]
                    if len(pair) == 2:
                        atom_pairs.append(pair)
                    else:
                        self.setting_error(
                            "TDISTANCE is incorrectly specified.")
                if len(atom_pairs) > 0:
                    self.set_parameter('tdistance', atom_pairs)

            # Projection direction used for thermal displacements and PDOS
            if conf_key == 'projection_direction':
                vals = [float(x) for x in confs['projection_direction'].split()]
                if len(vals) < 3:
                    self.setting_error(
                        "PROJECTION_DIRECTION (--pd) is incorrectly specified.")
                else:
                    self.set_parameter('projection_direction', vals)

            # Group velocity
            if conf_key == 'group_velocity':
                if confs['group_velocity'].lower() == '.true.':
                    self.set_parameter('is_group_velocity', True)

            # Moment of phonon states distribution
            if conf_key == 'moment':
                if confs['moment'].lower() == '.true.':
                    self.set_parameter('moment', True)

            if conf_key == 'moment_order':
                self.set_parameter('moment_order', int(confs['moment_order']))

            # Number of supercells with random displacements
            if conf_key == 'random_displacements':
                self.set_parameter('random_displacements',
                                   int(confs['random_displacements']))

            if conf_key == 'random_seed':
                self.set_parameter('random_seed', int(confs['random_seed']))

            # Use Lapack solver via Lapacke
            if conf_key == 'lapack_solver':
                if confs['lapack_solver'].lower() == '.true.':
                    self.set_parameter('lapack_solver', True)

    def _parse_conf_modulation(self, conf_modulation):
        modulation = {}
        modulation['dimension'] = [1, 1, 1]
        modulation['order'] = None
        mod_list = conf_modulation.split(',')
        header = mod_list[0].split()
        if len(header) > 2 and len(mod_list) > 1:
            if len(header) > 8:
                dimension = [int(x) for x in header[:9]]
                modulation['dimension'] = dimension
                if len(header) > 11:
                    delta_q = [float(x) for x in header[9:12]]
                    modulation['delta_q'] = delta_q
                if len(header) == 13:
                    modulation['order'] = int(header[12])
            else:
                dimension = [int(x) for x in header[:3]]
                modulation['dimension'] = dimension
                if len(header) > 3:
                    delta_q = [float(x) for x in header[3:6]]
                    modulation['delta_q'] = delta_q
                if len(header) == 7:
                    modulation['order'] = int(header[6])

            vals = []
            for phonon_mode in mod_list[1:]:
                mode_conf = [x for x in phonon_mode.split()]
                if len(mode_conf) < 4 or len(mode_conf) > 6:
                    self.setting_error("MODULATION tag is wrongly set.")
                    break
                else:
                    q = [fracval(x) for x in mode_conf[:3]]

                if len(mode_conf) == 4:
                    vals.append([q, int(mode_conf[3]) - 1, 1.0, 0])
                elif len(mode_conf) == 5:
                    vals.append([q,
                                 int(mode_conf[3]) - 1,
                                 float(mode_conf[4]),
                                 0])
                else:
                    vals.append([q,
                                 int(mode_conf[3]) - 1,
                                 float(mode_conf[4]),
                                 float(mode_conf[5])])

            modulation['modulations'] = vals
            self.set_parameter('modulation', modulation)
        else:
            self.setting_error("MODULATION tag is wrongly set.")

    def _set_settings(self):
        self.set_settings()
        params = self._parameters

        # Is getting least displacements?
        if 'create_displacements' in params:
            if params['create_displacements']:
                self._settings.set_run_mode('displacements')

        # Is force constants written or read?
        if 'force_constants' in params:
            if params['force_constants'] == 'write':
                self._settings.set_write_force_constants(True)
            elif params['force_constants'] == 'read':
                self._settings.set_read_force_constants(True)

        if 'read_force_constants' in params:
            self._settings.set_read_force_constants(
                params['read_force_constants'])

        if 'write_force_constants' in params:
            self._settings.set_write_force_constants(
                params['write_force_constants'])

        if 'is_full_fc' in params:
            self._settings.set_is_full_fc(params['is_full_fc'])

        # Enforce space group symmetyr to force constants?
        if 'fc_spg_symmetry' in params:
            self._settings.set_fc_spg_symmetry(params['fc_spg_symmetry'])

        if 'readfc_format' in params:
            self._settings.set_readfc_format(params['readfc_format'])

        if 'writefc_format' in params:
            self._settings.set_writefc_format(params['writefc_format'])

        # Use hdf5?
        if 'hdf5' in params:
            self._settings.set_is_hdf5(params['hdf5'])

        # Cutoff radius of force constants
        if 'cutoff_radius' in params:
            self._settings.set_cutoff_radius(params['cutoff_radius'])

        # band & mesh mode
        # This has to come before 'mesh_numbers' and 'band_paths'
        if 'mesh_numbers' in params and 'band_paths' in params:
            self._settings.set_run_mode('band_mesh')

        # Mesh
        if 'mesh_numbers' in params:
            if self._settings.get_run_mode() != 'band_mesh':
                self._settings.set_run_mode('mesh')
            self._settings.set_mesh_numbers(params['mesh_numbers'])
        if (self._settings.get_run_mode() == 'mesh' or
            self._settings.get_run_mode() == 'band_mesh'):
            if 'mp_shift' in params:
                self._settings.set_mesh_shift(params['mp_shift'])
            if 'is_time_reversal_symmetry' in params:
                if not params['is_time_reversal_symmetry']:
                    self._settings.set_time_reversal_symmetry(False)
            if 'is_mesh_symmetry' in params:
                if not params['is_mesh_symmetry']:
                    self._settings.set_is_mesh_symmetry(False)
            if 'is_gamma_center' in params:
                if params['is_gamma_center']:
                    self._settings.set_is_gamma_center(True)
            if 'mesh_format' in params:
                self._settings.set_mesh_format(params['mesh_format'])

        # band mode
        if 'band_paths' in params:
            if self._settings.get_run_mode() != 'band_mesh':
                self._settings.set_run_mode('band')
        if (self._settings.get_run_mode() == 'band' or
            self._settings.get_run_mode() == 'band_mesh'):
            if 'band_format' in params:
                self._settings.set_band_format(params['band_format'])
            if 'band_labels' in params:
                self._settings.set_band_labels(params['band_labels'])
            if 'band_connection' in params:
                self._settings.set_is_band_connection(params['band_connection'])
            if 'legacy_plot' in params:
                self._settings.set_is_legacy_plot(params['legacy_plot'])

        # Q-points mode
        if 'qpoints' in params or 'read_qpoints' in params:
            self._settings.set_run_mode('qpoints')
        if self._settings.get_run_mode() == 'qpoints':
            if 'qpoints_format' in params:
                self._settings.set_qpoints_format(params['qpoints_format'])

        # Whether write out dynamical matrices or not
        if 'write_dynamical_matrices' in params:
            if params['write_dynamical_matrices']:
                self._settings.set_write_dynamical_matrices(True)

        # Whether write out mesh.yaml or mesh.hdf5
        if 'write_mesh' in params:
            self._settings.set_write_mesh(params['write_mesh'])

        # Anime mode
        if 'anime_type' in params:
            self._settings.set_anime_type(params['anime_type'])

        if 'anime' in params:
            self._settings.set_run_mode('anime')
            anime_type = self._settings.get_anime_type()
            if anime_type == 'v_sim':
                qpoints = [fracval(x) for x in params['anime'][0:3]]
                self._settings.set_anime_qpoint(qpoints)
                if len(params['anime']) > 3:
                    self._settings.set_anime_amplitude(float(params['anime'][3]))
            else:
                self._settings.set_anime_band_index(int(params['anime'][0]))
                self._settings.set_anime_amplitude(float(params['anime'][1]))
                self._settings.set_anime_division(int(params['anime'][2]))
            if len(params['anime']) == 6:
                self._settings.set_anime_shift(
                    [fracval(x) for x in params['anime'][3:6]])

        # Modulation mode
        if 'modulation' in params:
            self._settings.set_run_mode('modulation')
            self._settings.set_modulation(params['modulation'])

        # Character table mode
        if 'irreps_qpoint' in params:
            self._settings.set_run_mode('irreps')
            self._settings.set_irreps_q_point(
                params['irreps_qpoint'][:3])
            if len(params['irreps_qpoint']) == 4:
                self._settings.set_irreps_tolerance(params['irreps_qpoint'][3])
            if self._settings.get_run_mode() == 'irreps':
                if 'show_irreps' in params:
                    self._settings.set_show_irreps(params['show_irreps'])
                if 'little_cogroup' in params:
                    self._settings.set_is_little_cogroup(params['little_cogroup'])

        # DOS
        if 'dos_range' in params:
            fmin =  params['dos_range'][0]
            fmax =  params['dos_range'][1]
            fpitch = params['dos_range'][2]
            self._settings.set_min_frequency(fmin)
            self._settings.set_max_frequency(fmax)
            self._settings.set_frequency_pitch(fpitch)
        if 'dos' in params:
            self._settings.set_is_dos_mode(params['dos'])

        if 'fits_debye_model' in params:
            self._settings.set_fits_Debye_model(params['fits_debye_model'])

        if 'fmax' in params:
            self._settings.set_max_frequency(params['fmax'])

        if 'fmin' in params:
            self._settings.set_min_frequency(params['fmin'])

        # Project PDOS x, y, z directions in Cartesian coordinates
        if 'xyz_projection' in params:
            self._settings.set_xyz_projection(params['xyz_projection'])
            if ('pdos' not in params and
                self._settings.get_pdos_indices() is None):
                self.set_parameter('pdos', [])

        if 'pdos' in params:
            self._settings.set_pdos_indices(params['pdos'])
            self._settings.set_is_eigenvectors(True)
            self._settings.set_is_mesh_symmetry(False)

        if ('projection_direction' in params and
            not self._settings.get_xyz_projection()):
            self._settings.set_projection_direction(
                params['projection_direction'])
            self._settings.set_is_eigenvectors(True)
            self._settings.set_is_mesh_symmetry(False)

        # Thermal properties
        if 'tprop' in params:
            self._settings.set_is_thermal_properties(params['tprop'])
            # Exclusive conditions
            self._settings.set_is_thermal_displacements(False)
            self._settings.set_is_thermal_displacement_matrices(False)
            self._settings.set_is_thermal_distances(False)

        # Projected thermal properties
        if 'ptprop' in params and params['ptprop']:
            self._settings.set_is_thermal_properties(True)
            self._settings.set_is_projected_thermal_properties(True)
            self._settings.set_is_eigenvectors(True)
            self._settings.set_is_mesh_symmetry(False)
            # Exclusive conditions
            self._settings.set_is_thermal_displacements(False)
            self._settings.set_is_thermal_displacement_matrices(False)
            self._settings.set_is_thermal_distances(False)

        # Use imaginary frequency as real for thermal property calculation
        if 'pretend_real' in params:
            self._settings.set_pretend_real(params['pretend_real'])

        # Thermal displacements
        if 'tdisp' in params and params['tdisp']:
            self._settings.set_is_thermal_displacements(True)
            self._settings.set_is_eigenvectors(True)
            self._settings.set_is_mesh_symmetry(False)
            # Exclusive conditions
            self._settings.set_is_thermal_properties(False)
            self._settings.set_is_thermal_displacement_matrices(False)
            self._settings.set_is_thermal_distances(True)

        # Thermal displacement matrices
        if ('tdispmat' in params and params['tdispmat'] or
            'tdispmat_cif' in params):
            self._settings.set_is_thermal_displacement_matrices(True)
            self._settings.set_is_eigenvectors(True)
            self._settings.set_is_mesh_symmetry(False)
            # Exclusive conditions
            self._settings.set_is_thermal_properties(False)
            self._settings.set_is_thermal_displacements(False)
            self._settings.set_is_thermal_distances(False)

            # Temperature used to calculate thermal displacement matrix
            # to write aniso_U to cif
            if 'tdispmat_cif' in params:
                self._settings.set_thermal_displacement_matrix_temperature(
                    params['tdispmat_cif'])

        # Thermal distances
        if 'tdistance' in params:
            self._settings.set_is_thermal_distances(True)
            self._settings.set_is_eigenvectors(True)
            self._settings.set_is_mesh_symmetry(False)
            self._settings.set_thermal_atom_pairs(params['tdistance'])
            # Exclusive conditions
            self._settings.set_is_thermal_properties(False)
            self._settings.set_is_thermal_displacements(False)
            self._settings.set_is_thermal_displacement_matrices(False)

        # Group velocity
        if 'is_group_velocity' in params:
            self._settings.set_is_group_velocity(params['is_group_velocity'])

        # Moment mode
        if 'moment' in params:
            self._settings.set_is_moment(params['moment'])
            self._settings.set_is_eigenvectors(True)
            self._settings.set_is_mesh_symmetry(False)

        if self._settings.get_is_moment():
            if 'moment_order' in params:
                self._settings.set_moment_order(params['moment_order'])

        # Number of supercells with random displacements
        if 'random_displacements' in params:
            self._settings.set_random_displacements(
                params['random_displacements'])

        if 'random_seed' in params:
            self._settings.set_random_seed(params['random_seed'])

        # Use Lapack solver via Lapacke
        if 'lapack_solver' in params:
            self._settings.set_lapack_solver(params['lapack_solver'])



# Copyright (C) 2018 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys
import numpy as np


def get_fc2(supercell,
            primitive,
            displacements,
            forces,
            atom_list=None,
            options=None,
            log_level=0):
    p2s_map = primitive.p2s_map
    is_compact_fc = (atom_list is not None and
                     (atom_list == p2s_map).all())
    fc2 = run_alm(supercell,
                  primitive,
                  displacements,
                  forces,
                  1,
                  is_compact_fc=is_compact_fc,
                  options=options,
                  log_level=log_level)[0]

    if not is_compact_fc and atom_list is not None:
        fc2 = np.array(fc2[atom_list], dtype='double', order='C')

    return fc2


def run_alm(supercell,
            primitive,
            displacements,
            forces,
            maxorder,
            is_compact_fc,
            options=None,
            log_level=0):
    fcs = None  # This is returned.

    lattice = supercell.cell
    positions = supercell.scaled_positions
    numbers = supercell.numbers
    natom = len(numbers)
    p2s_map = primitive.p2s_map
    p2p_map = primitive.p2p_map

    alm_options = _update_options(options)
    num_elems = len(np.unique(numbers))
    if log_level:
        print("---------------------------------"
              " ALM start "
              "--------------------------------")
        print("ALM is a non-trivial force constants calculator. "
              "Please cite the paper:")
        print("T. Tadano and S. Tsuneyuki, "
              "J. Phys. Soc. Jpn. 87, 041015 (2018).")
        print("ALM is developed at https://github.com/ttadano/ALM by T. "
              "Tadano.")
    if log_level == 1:
        print("Use -v option to watch detailed ALM log.")

    shape = (maxorder, num_elems, num_elems)
    cutoff_radii = -np.ones(shape, dtype='double')
    if alm_options['cutoff'] is not None:
        if len(alm_options['cutoff']) == 1:
            cutoff_radii[:] = alm_options['cutoff'][0]
        elif np.prod(shape) == len(alm_options['cutoff']):
            cutoff_radii[:] = np.reshape(alm_options['cutoff'], shape)
        else:
            raise RuntimeError("Cutoff is not properly set.")

    _disps, _forces, df_msg = _slice_displacements_and_forces(
        displacements,
        forces,
        alm_options['ndata'],
        alm_options['nstart'],
        alm_options['nend'])

    if log_level > 0:
        log_level_alm = log_level - 1
    else:
        log_level_alm = 0

    try:
        from alm import ALM, optimizer_control_data_types
    except ImportError:
        raise ImportError("ALM python module was not found.")

    with ALM(lattice, positions, numbers) as alm:
        if log_level > 0:
            if alm_options['cutoff'] is not None:
                for i in range(maxorder):
                    if maxorder > 1:
                        print("fc%d" % (i + 2))
                    print(("cutoff" + " %6s" * num_elems)
                          % tuple(alm.kind_names.values()))
                    for r, kn in zip(cutoff_radii[i], alm.kind_names.values()):
                        print(("   %-3s" + " %6.2f" * num_elems)
                              % ((kn, ) + tuple(r)))
            if df_msg is not None:
                print(df_msg)
        if log_level > 1:
            print("")
        sys.stdout.flush()

        alm.output_filename_prefix = alm_options['output_filename_prefix']
        alm.verbosity = log_level_alm

        alm.define(
            maxorder,
            cutoff_radii=cutoff_radii,
            symmetrization_basis=alm_options['symmetrization_basis'])
        alm.displacements = _disps
        alm.forces = _forces

        # Mainly for elastic net (or lasso) regression
        optcontrol = {}
        for key in optimizer_control_data_types:
            if key in alm_options:
                optcontrol[key] = alm_options[key]
        if optcontrol:
            alm.optimizer_control = optcontrol
            if ('cross_validation' in optcontrol and
                optcontrol['cross_validation'] > 0):
                alm.optimize(solver=alm_options['solver'])
                optcontrol['cross_validation'] = 0
                optcontrol['l1_alpha'] = alm.cv_l1_alpha
                alm.optimizer_control = optcontrol

        alm.optimize(solver=alm_options['solver'])

        fcs = _extract_fc_from_alm(alm,
                                   natom,
                                   maxorder,
                                   is_compact_fc,
                                   p2s_map=p2s_map,
                                   p2p_map=p2p_map)

    if log_level:
        print("----------------------------------"
              " ALM end "
              "---------------------------------")

    return fcs


def _update_options(fc_calculator_options):
    """Set ALM options with appropriate data types

    fc_calculator_options : str
        This string should be written such as follows:

            "solver = dense, cutoff = 5"

        This string is parsed as collection of settings that are separated by
        comma ','. Each setting has the format of 'option = value'. The value
        is cast to have its appropriate data type for ALM in this method.

    """

    try:
        from alm import optimizer_control_data_types
    except ImportError:
        raise ImportError("ALM python module was not found.")

    # Default settings.
    alm_options = {'solver': 'SimplicialLDLT',
                   'ndata': None,
                   'nstart': None,
                   'nend': None,
                   'cutoff': None,
                   'symmetrization_basis': 'Lattice',
                   'output_filename_prefix': None}

    if fc_calculator_options is not None:
        alm_option_types = {'cutoff': np.double,
                            'ndata': int,
                            'nstart': int,
                            'nend': int,
                            'output_filename_prefix': str,
                            'solver': str,
                            'symmetrization_basis': str}
        alm_option_types.update(optimizer_control_data_types)
        for option_str in fc_calculator_options.split(","):
            key, val = [x.strip() for x in option_str.split('=')[:2]]
            if key.lower() in alm_option_types:
                if alm_option_types[key.lower()] is np.double:
                    option_value = np.array(
                        [float(x) for x in val.split()], dtype='double')
                elif alm_option_types[key.lower()] is np.intc:
                    option_value = np.array(
                        [int(x) for x in val.split()], dtype='intc')
                else:
                    option_value = alm_option_types[key.lower()](val)
                alm_options[key] = option_value
    return alm_options


def _slice_displacements_and_forces(d, f, ndata, nstart, nend):
    msg = None
    if ndata is not None:
        _d = d[:ndata]
        _f = f[:ndata]
        msg = "Number of displacement supercells: %d" % ndata
    elif nstart is not None and nend is not None:
        _d = d[nstart - 1:nend]
        _f = f[nstart - 1:nend]
        msg = "Supercell index range: %d - %d" % (nstart, nend)
    else:
        return d, f, None

    return (np.array(_d, dtype='double', order='C'),
            np.array(_f, dtype='double', order='C'), msg)


def _extract_fc_from_alm(alm,
                         natom,
                         maxorder,
                         is_compact_fc,
                         p2s_map=None,
                         p2p_map=None):
    # Harmonic: order=1, 3rd: order=2, ...
    fcs = []
    for order in range(1, maxorder + 1):
        fc = None
        p2s_map_alm = alm.getmap_primitive_to_supercell()[0]
        if (is_compact_fc and
            len(p2s_map_alm) == len(p2s_map) and
            (p2s_map_alm == p2s_map).all()):
            fc_shape = (
                (len(p2s_map), ) + (natom, ) * order + (3, ) * (order + 1))
            fc = np.zeros(fc_shape, dtype='double', order='C')
            for fc_elem, indices in zip(
                    *alm.get_fc(order, mode='origin', permutation=1)):
                v = indices // 3
                c = indices % 3
                selection = (p2p_map[v[0]], ) + tuple(v[1:]) + tuple(c)
                fc[selection] = fc_elem

        if fc is None:
            if is_compact_fc:
                atom_list = p2s_map
            else:
                atom_list = np.arange(natom, dtype=int)
            fc_shape = (
                (len(atom_list), ) + (natom, ) * order + (3, ) * (order + 1))
            fc = np.zeros(fc_shape, dtype='double', order='C')
            for fc_elem, indices in zip(
                    *alm.get_fc(order, mode='all', permutation=1)):
                v = indices // 3
                idx = np.where(atom_list == v[0])[0]
                if len(idx) > 0:
                    c = indices % 3
                    selection = (idx[0], ) + tuple(v[1:]) + tuple(c)
                    fc[selection] = fc_elem

        fcs.append(fc)

    return fcs



# Copyright (C) 2015 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from phonopy.phonon.thermal_properties import ThermalProperties

class GruneisenThermalProperties(object):
    def __init__(self,
                 gruneisen_mesh,
                 volumes,
                 t_step=2,
                 t_max=2004,
                 t_min=0,
                 cutoff_frequency=None):
        phonon = gruneisen_mesh.get_phonon()
        self._cutoff_frequency = cutoff_frequency
        self._factor = phonon.get_unit_conversion_factor(),
        self._V0 = phonon.get_primitive().get_volume()
        self._gamma = gruneisen_mesh.get_gruneisen()
        self._gamma_prime = gruneisen_mesh.get_gamma_prime()
        self._weights = gruneisen_mesh.get_weights()
        self._eigenvalues = gruneisen_mesh.get_eigenvalues()
        self._frequencies = gruneisen_mesh.get_frequencies()
        self._thermal_properties = []
        for V in volumes:
            tp = self._get_thermal_properties_at_V(V)
            tp.set_temperature_range(t_min=t_min, t_max=t_max, t_step=t_step)
            tp.run()
            self._thermal_properties.append(tp)

    def get_thermal_properties(self):
        """Return a set of phonopy.phonon::ThermalProperties object"""
        return self._thermal_properties

    def write_yaml(self, filename='thermal_properties'):
        for i, tp in enumerate(self._thermal_properties):
            tp.write_yaml(filename="%s-%02d.yaml" % (filename, i))

    def _get_thermal_properties_at_V(self, V):
        frequencies = self._get_frequencies_at_V(V)
        tp = ThermalProperties(frequencies,
                               weights=self._weights,
                               cutoff_frequency=self._cutoff_frequency)
        return tp

    def _get_frequencies_at_V(self, V):
        return self._get_frequencies_at_V_analytical_solution(V)

    def _get_frequencies_at_V_analytical_solution(self, V):
        eigvals = self._eigenvalues * np.exp(-2 * self._gamma *
                                             np.log(V / self._V0))
        return np.sqrt(abs(eigvals)) * np.sign(eigvals) * self._factor

    def _get_frequencies_at_V_analytical_solution_with_1st_correction(self, V):
        V0 = self._V0
        g_prime = self._gamma_prime / V0
        eigvals = self._eigenvalues * np.exp(
            -2 * ((self._gamma - g_prime * V0) * np.log(V / V0)
                  + g_prime * (V - V0)))
        return np.sqrt(abs(eigvals)) * np.sign(eigvals) * self._factor

    def _get_frequencies_at_V_Taylor_expansion_to_1st_order(self, V):
        return self._frequencies * (
            1.0
            - self._gamma * (V - self._V0) / self._V0)

    def _get_frequencies_at_V_Taylor_expansion_to_2nd_order(self, V):
        return self._frequencies * (
            1.0
            - self._gamma * (V - self._V0) / self._V0
            - self._gamma_prime * ((V - self._V0) / self._V0) ** 2 / 2)



import unittest
import os
import numpy as np
import phonopy

data_dir = os.path.dirname(os.path.abspath(__file__))

tp_str = """0.000000 100.000000 200.000000 300.000000 400.000000
500.000000 600.000000 700.000000 800.000000 900.000000
4.856373 3.916036 -0.276031 -6.809284 -14.961974
-24.342086 -34.708562 -45.898943 -57.796507 -70.313405
0.000000 26.328820 55.258118 74.269718 88.156943
99.054231 108.009551 115.606163 122.200204 128.024541
0.000000 36.207859 45.673598 47.838871 48.634251
49.009334 49.214960 49.339592 49.420745 49.476502"""


class TestThermalProperties(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testThermalProperties(self):
        phonon = self._get_phonon()
        phonon.run_mesh([5, 5, 5])
        phonon.run_thermal_properties(t_step=100, t_max=900)
        tp = phonon.thermal_properties
        np.testing.assert_allclose(tp.temperatures, tp.get_temperatures())
        np.testing.assert_allclose(tp.thermal_properties,
                                   tp.get_thermal_properties())
        np.testing.assert_allclose(tp.zero_point_energy,
                                   tp.get_zero_point_energy())
        np.testing.assert_allclose(tp.high_T_entropy,
                                   tp.get_high_T_entropy())
        np.testing.assert_allclose(tp.number_of_integrated_modes,
                                   tp.get_number_of_integrated_modes())
        np.testing.assert_allclose(tp.number_of_modes,
                                   tp.get_number_of_modes())

        tp_ref = np.reshape([float(x) for x in tp_str.split()], (-1, 10))
        np.testing.assert_allclose(tp.thermal_properties, tp_ref, atol=1e-5)

    def _get_phonon(self):
        phonon = phonopy.load(
            supercell_matrix=[[2, 0, 0], [0, 2, 0], [0, 0, 2]],
            primitive_matrix=[[0, 0.5, 0.5],
                              [0.5, 0, 0.5],
                              [0.5, 0.5, 0]],
            unitcell_filename=os.path.join(data_dir, "..", "POSCAR_NaCl"),
            force_sets_filename=os.path.join(data_dir, "..",
                                             "FORCE_SETS_NaCl"),
            born_filename=os.path.join(data_dir, "..", "BORN_NaCl"))
        return phonon


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestThermalProperties)
    unittest.TextTestRunner(verbosity=2).run(suite)



from phonopy import Phonopy
from phonopy.interface.vasp import read_vasp
from phonopy.file_IO import parse_FORCE_SETS
import numpy as np

def get_frequency(poscar_filename, force_sets_filename):
    unitcell = read_vasp(poscar_filename)
    volume = unitcell.get_volume()
    phonon = Phonopy(unitcell,
                     [[2, 0, 0],
                      [0, 2, 0],
                      [0, 0, 2]],
                     primitive_matrix=[[0, 0.5, 0.5],
                                       [0.5, 0, 0.5],
                                       [0.5, 0.5, 0]])
    force_sets = parse_FORCE_SETS(filename=force_sets_filename)
    phonon.set_displacement_dataset(force_sets)
    phonon.produce_force_constants()
    return phonon.get_frequencies([0.5, 0.5, 0]), volume

frequencies = []
volumes = []
for i in range(-10, 6):
    poscar_filename = "POSCAR-%d" % i
    force_sets_filename = "FORCE_SETS-%d" % i
    fs, v = get_frequency(poscar_filename, force_sets_filename)
    frequencies.append(fs)
    volumes.append(v)

import matplotlib.pyplot as plt
for freq_at_X in np.array(frequencies).T:
    freq_squared = freq_at_X ** 2 * np.sign(freq_at_X)
    # np.sign is used to treat imaginary mode, since
    # imaginary frequency is returned as a negative value from phonopy.
    plt.plot(volumes, freq_squared, "o-")
    # for v, f2 in zip(volumes, freq_squared):
    #      print("%f %f" % (v, f2))
    # print('')
    # print('')
plt.title("Frequeny squared (THz^2) at X-point vs volume (A^3)")
plt.show()



# Copyright (C) 2012 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np


def get_eos(eos):
    # Third-order Birch-Murnaghan EOS
    def birch_murnaghan(v, *p):
        """
        p[0] = E_0
        p[1] = B_0
        p[2] = B'_0
        p[3] = V_0
        """
        return p[0] + 9.0 / 16 * p[3] * p[1] * (
            ((p[3] / v)**(2.0 / 3) - 1)**3 * p[2] +
            ((p[3] / v)**(2.0 / 3) - 1)**2 * (6 - 4 * (p[3] / v)**(2.0 / 3)))

    # Murnaghan EOS
    def murnaghan(v, *p):
        """
        p[0] = E_0
        p[1] = B_0
        p[2] = B'_0
        p[3] = V_0
        """
        return (p[0]
                + p[1] * v / p[2] *((p[3] / v)**p[2] / (p[2] - 1) + 1)
                - p[1] * p[3] / (p[2] - 1))

    # Vinet EOS
    def vinet(v, *p):
        """
        p[0] = E_0
        p[1] = B_0
        p[2] = B'_0
        p[3] = V_0
        """

        x = (v / p[3]) ** (1.0 / 3)
        xi = 3.0 / 2 * (p[2] - 1)
        return p[0] + (9 * p[1] * p[3] / (xi**2)
                       * (1 + (xi * (1 - x) - 1) * np.exp(xi * (1 - x))))

    if eos == 'murnaghan':
        return murnaghan
    elif eos == 'birch_murnaghan':
        return birch_murnaghan
    else:
        return vinet


def fit_to_eos(volumes, fe, eos):
    fit = EOSFit(volumes, fe, eos)
    fit.fit([fe[len(fe) // 2], 1.0, 4.0, volumes[len(volumes) // 2]])

    return fit.parameters


class EOSFit(object):
    """

    Attributes
    ----------
    parameters: ndarray
        Fitting parameters to EOS corresponding to [energy, B, B', V].
        dtype=float
        shape=(4,)

    """

    def __init__(self, volume, energy, eos):
        self._energy = np.array(energy)
        self._volume = np.array(volume)
        self._eos = eos

        self.parameters = None

    def fit(self, initial_parameters):
        import sys
        import logging
        import warnings

        try:
            from scipy.optimize import leastsq
            import scipy
        except ImportError:
            print("You need to install python-scipy.")
            sys.exit(1)

        warnings.filterwarnings('error')

        def residuals(p, eos, v, e):
            return eos(v, *p) - e

        try:
            result = leastsq(residuals,
                             initial_parameters,
                             args=(self._eos, self._volume, self._energy),
                             full_output=1)
        except RuntimeError:
            logging.exception('Fitting to EOS failed.')
            raise
        except (RuntimeWarning, scipy.optimize.optimize.OptimizeWarning):
            logging.exception('Difficulty in fitting to EOS.')
            raise
        else:
            self.parameters = result[0]
